;;;;  Copyright (c) 1992, Giuseppe Attardi.
;;;;
;;;;    This program is free software; you can redistribute it and/or
;;;;    modify it under the terms of the GNU Library General Public
;;;;    License as published by the Free Software Foundation; either
;;;;    version 2 of the License, or (at your option) any later version.
;;;;
;;;;    See file '../Copyright' for full details.
;;;;
;;;;  errors.lsp

(in-package "SYSTEM")

(use-package "CONDITIONS")

(setf (get 'universal-error-handler 'dbg::error-reporter) t)

#+Multithread
(defun universal-error-handler
    (error-name correctable function-name
		continue-format-string error-format-string
		args)
  (declare (ignore function-name continue-format-string))
  (declare (inline apply funcall));; So as not to get bogus frames in debugger
  ;;  Some conditions want special proceed options; they supply their own
  (if (or (plusp *break-level*) (not *scheduler-disabled-in-error*))
      (let ((signaller (get error-name 'CONDITION-SIGNALLER)))
	(if signaller
	    (apply signaller error-name error-format-string args)
	    (let ((condition
		   (or
		    (and (setq signaller (get error-name 'CONDITION-MAKER))
			 (funcall signaller
				  correctable
				  error-format-string args))
		    (get error-name 'CONDITION-NAME)
		    (default-condition-maker
			error-format-string args))))
	      (if correctable
		  (cerror continue-format-string condition)
		  (error condition)))))
      (unwind-protect
	   (let ((signaller (get error-name 'CONDITION-SIGNALLER)))
	     (sys:thread-break-in)
	     (if signaller
		 (apply signaller error-name error-format-string args)
		 (let ((condition
			(or
			 (and (setq signaller
				    (get error-name 'CONDITION-MAKER))
			      (funcall signaller
				       correctable
				       error-format-string args))
			 (get error-name 'CONDITION-NAME)
			 (default-condition-maker
			     error-format-string args))))
		   (if correctable
		       (cerror continue-format-string condition)
		       (error condition))))) ;always hit debugger.
	(sys:thread-break-resume))))

#-Multithread
(defun universal-error-handler
    (error-name correctable function-name
		continue-format-string error-format-string
		args &aux thing)
  (declare (ignore function-name continue-format-string))
  (declare (inline apply funcall));; So as not to get bogus frames in debugger
  ;;  Some conditions want special proceed options; they supply their 
  (if (setq thing (get error-name 'CONDITION-SIGNALLER))
      (apply thing error-name error-format-string args)
      (let ((condition (or
			(and (setq thing (get error-name 'CONDITION-MAKER))
			     (funcall thing
				      correctable
				      error-format-string args))
			(get error-name 'CONDITION-NAME)
			(default-condition-maker
			    error-format-string args))))
	(if correctable
	    (cerror continue-format-string condition)
	    (error condition)))))	;always hit debugger.
    
(defun terminal-interrupt (correctablep)
  (flet ((term-interrupt-internal ()
           (invoke-debugger
	    (make-condition 'SIMPLE-CONDITION
			    :format-string "Terminal Interrupt"
			    :format-arguments ()))))
    (if correctablep
	(with-simple-restart (continue "Return from interrupt")
				 (term-interrupt-internal))
	  (term-interrupt-internal))))

(defun default-condition-maker (format-string args)
  (make-condition 'SIMPLE-ERROR
		  :format-string format-string
		  :format-arguments args))


(defmacro defun-property ((symbol indicator) args &body body)
  `(setf (get ',symbol ',indicator) #'(lambda ,args . ,body)))

(setf (get '|(PROPERTY :UNBOUND-VARIABLE CONDITION-SIGNALLER)|
	   'dbg:error-reporter)
      t)

(defun-property (:unbound-variable condition-signaller)
         (error-name fs variable &rest args)
  (declare (ignore error-name fs args))
  (restart-case (error 'UNBOUND-VARIABLE :name variable)
    (store-value (value)
      :report (lambda (st)
		(format st "Provide a value to store as the value of ~s"
			variable))
      :interactive  (lambda ()
		     (format t "~&Form to evaluate for new value of ~s:"
			     variable)
		     (list (eval (read))))
      (setf (symbol-value variable) value))
    (continue ()
      :report (lambda (st)
		(format st "Retry getting the value of ~s" variable))
      nil)))


(setf (get '|(PROPERTY :UNDEFINED-FUNCTION CONDITION-SIGNALLER)|
	   'dbg:error-reporter)
      t)

(defun-property (:undefined-function condition-signaller)
  		(error-name fs function &rest args)
  (declare (ignore error-name fs args))
  (restart-case (error 'UNDEFINED-FUNCTION :name function)
    (store-value (value)
      :report (lambda (st)
		(format st "Provide a value to store as the definition of ~s"
			function))
      :interactive (lambda ()
		     (format t "~&Form to evaluate for new definition of ~s:"
			     function)
		     (list (eval (read))))
      (setf (symbol-function function) value))
    (continue ()
      :report (lambda (st)
		(format st "Retry getting the definition of ~s" function))
      nil)))
      
;;; ----------------------------------------------------------------------
;;; Debugger interface

;(defvar *debug-level* 0) ; use *break-level* instead. Beppe
(defvar *condition*)
(defvar *base-frame* 2)
(defvar *initial-frame*)
(defvar *top-frame*)
(defvar *current-frame*)
(defvar *current-environment*)

#| This should replace item in break-commands:
      ((:c :continue) tpl-proceed-command nil
       ":c(ontinue)	Continue execution with proceed option"
       ":continue &optional (n 0)			[Break command]~@
	:c &optional (n 0)				[Abbreviation]~@
	~@
	Continue execution with selected proceed option.~%")
|#

(defun tpl-continue-command (&optional (opt-num 0))
  (let ((pcases (compute-restarts)))
    (if (< -1 opt-num (length pcases))
      (invoke-restart-interactively (nth opt-num pcases))
      (format t "~% There are ~[no~:;only ~:*~d~] restarts." 
	      (length pcases)))))

(defun invoke-debugger (condition)
  ;;  CONDITION really has to be a condition object.
  (with-debugger-environment (condition)
    (set-current-frame *current-frame*) ; to get environment right
    (print-initial-debugger-message) ; >>Error: <report> \ proc options
    (break-level t "")))

(defmacro with-debugger-environment ((condition) &body body)
  `(let* ((*standard-input* *debug-io*)
	  (*standard-output* *debug-io*)
	  (*error-output* *debug-io*)
	  (*print-escape* nil)
	  (*print-pretty* t)
	  (*condition* ,condition)
;	  (*debug-level* (1+ *debug-level*))
	  (*top-frame* (- (ihs-top) 2))
	  (*initial-frame* (skip-error-reporter-frames *top-frame*))
	  (*current-frame* *initial-frame*)
	  (*current-environment* nil))
     . ,body))

(defun skip-error-reporter-frames (frm)
  (loop
    (when (zerop frm) (return frm))
       (unless (or (get (ihs-fname frm) 'ERROR-REPORTER)
		   (not (ihs-visible frm)))
	 (return frm))
       (decf frm)))

(defun set-current-frame  (frame-number)
  (setq *current-frame* frame-number
	*current-environment*
	(when (typep (ihs-fun frame-number)
		     '(OR SYMBOL LIST))
	  (let ((i (ihs-vs frame-number)))
	    (list (vs i) (vs (1+ i)) (vs (+ 2 i)))))))

#| not used. Beppe
(defun debugger-command-level ( &aux (level *debug-level*) back-p)
  (catch 'DEBUGGER-ABORT      ; :Q command throws to this.
    (loop
     (when back-p
       (paranoid-format t "~%Back to Debugger Level ~d.~%" level))
     (setq back-p t)  ; until proven otherwise...
     (with-simple-restart
            (abort "Debugger Level ~d" level)
      ;; prompt...
      (terpri)
      (unless (eq *package* (find-package "USER"))
        (princ (package-name *package*)))
      (write-string " -")
      (dotimes (i level) (write-char #\>))
      ;; Read...
      ;;  (preserve whitespace so the read-line below always works)
      (let ((command (read-preserving-whitespace 
		      *standard-input* nil sys::*eof*))
	    cmd-fcn)
      	;; Execute...
      	(cond
	 ((eq command sys::*eof*)  (bye))	; Exit nicely...
       	 ((and (keywordp command)		; :X
	       (setq cmd-fcn (get command 'DEBUGGER-COMMAND)))
	  ;;  Non-parenthetical style debugger command.
	  ;;  Gather up trailing args, e.g. ":proc 3"
	  (let (arglist 
		(line (read-line))
		(start 0))
	    (loop (multiple-value-bind (thing new-start)
		       (read-from-string line nil sys::*eof* :start start)
		    (setq start new-start)
		    (when (eq thing sys::*eof*) (return))
		    (push thing arglist)))
	    (setq arglist (nreverse arglist))
	    (apply cmd-fcn arglist)))
       	 (t			; (+ 1 2) etc.
	  (command-level-eval-print command))))
       ;; Haven't taken the ABORT restart, so don't say "Back to..."
       (setq back-p nil))))
  ;; Only way to get here is to say :Q.  Thus, try ABORT restart.
  ;; If no ABORT catchers, just return.  This SHOULD never happen, but...
  (abort)
  (format *error-output* "Debugger: no ABORT restarts in :Q"))
|#

(defun print-initial-debugger-message (&optional skip-first-restart-option)
  (let ((*print-escape* nil))
    (paranoid-format t "~&>>~a: ~a"
		     (etypecase *condition*
		       (error "Error")
		       (break "Break")
		       (warning "Warning")
		       (condition "Condition"))
	    *condition*)
    (print-function-and-arguments)
    (print-restart-options skip-first-restart-option)
    ))

(defun print-brief-frame-info (&optional (frame *current-frame*))
  (print-function-and-arguments frame)
  (when (special-bindings-for-frame-p frame)
	(format t "~% There are special variables bound."))
  (when (catch-blocks-for-frame-p frame)
	(format t "~% There are CATCH, BLOCK, and/or TAGBODY tags.")))

(defun print-function-and-arguments (&optional (frame *current-frame*))
  (let ((fun (ihs-fun frame))
	(fname (ihs-fname frame)))
    (cond
     ((and (consp fun)
	   (eq (car fun) 'LAMBDA-CLOSURE))
      (paranoid-format t "~% an internal of ~s"
		       (first (first (fourth fun)))))
     ((and (consp fun)
	   (eq (car fun) 'LAMBDA))
      (paranoid-format t "~% ~s" fun))
     (t
	(paranoid-format t "~% ~s" fname))))
  (paranoid-format t "~30t(IHS[~d])" frame)
  (print-frame-arguments frame))

(defun print-frame-arguments (&optional (frame *current-frame*))
  (let ((func (ihs-fun frame)))
    (typecase func
      (compiled-function
	(print-stack-info-for-frame frame))
      (symbol				; Special operator, e.g. LET, PROGN...
        ())
      (list				;lambda-mumble
        (flet ((print-lambda-args (arglist)
	      (do* ((locals (vs (ihs-vs frame)))
		    (args arglist (cdr args))
		    (arg (car args) (car args))
		    (argnum 0 (1+ argnum))
		    (rest-arg-p))
		   ((endp args) nil)
		(case arg
		  ((&REST &BODY)  (setq rest-arg-p t))
		  (&OPTIONAL   (decf argnum))	; the &optional doesn't count.
		  (&KEY
		    (format t "~%<<keyword args>>"); handwave
		    (return nil))
		  (OTHERWISE
		   (if rest-arg-p
		       (paranoid-format t "~%Rest arg (~a): ~s"
			       arg (lambda-argument arg func locals))
		       (paranoid-format t "~%Arg ~d (~a): ~s"
			       argnum arg (lambda-argument arg func locals))))))))
	(print-lambda-args (lambda-parameter-list func))))
    (otherwise
      (paranoid-format t "~%What kind of a function is ~s?" func)))))

(defun value-stack-info (frame-number)
  (let ((low (ihs-vs frame-number))
	(nonarg (ihs-vstop frame-number))
	(top (if (<= (1+ frame-number) (ihs-top))
		(ihs-vs (1+ frame-number))
		(vs-top)))
	(arginfo (get (ihs-fname frame-number) 'SYS::ARGLIST)))
#| we keep arglist as single list. Beppe
    (when (third arginfo) ; rest arg
      (setq nonarg (min nonarg (+ low (length (first arginfo))
			               (length (second arginfo))))))
|#
    (values low (min top nonarg) top)))

;;; This prints the args and value stack for a compiled function frame.
(defun print-stack-info-for-frame (frame-number)
  (let ((*print-level* 2)	; Don't run away printing things...
	(*print-length* 3)
	(arglist (get (ihs-fname frame-number) 'SYS::ARGLIST)))
    (multiple-value-bind (low nonarg) (value-stack-info frame-number)
      (when (/= low nonarg)
	(do ((vsx low (1+ vsx))
	     (argn 0 (1+ argn))
	     (args arglist (cdr args))
	     (argname))
	    ((>= vsx nonarg))
	  (declare (fixnum vsx argn))
	  (setq argname (car args))
	  (if (member argname lambda-list-keywords :test #'eq)
	      (case argname
		(&REST
		 (return		; Nargs is false! Stop here!
		   (paranoid-format t "~%Rest arg~:[~;~:* (~a)~]: ~s"
				    argname (vs argn))))
		(&OPTIONAL
		 (format t "~% -- Optional args: --"))
		(&KEY)
		(OTHERWISE (return nil)))
	      (paranoid-format t "~%Arg ~d~:[~;~:* (~a)~]: ~s"
		      argn argname (vs vsx))))))))

;;; return the argument named <name> in the variables given.
(defun lambda-argument (name func variables)
  (when (member name (lambda-parameter-list func))
    (let* ((env (lambda-lexical-variables func))
	   (where (if env
		      (search env variables :test #'eql)
		    (length variables))))
      (second (find name variables
		    :start (max 0 (- where (lambda-parameter-count func)))
		    :end where
		    :test #'eql
		    :key #'car)))))	    
    

(defun lambda-parameter-list (func)
  (if (compiled-function-p func)
      (let ((name (sys::compiled-function-name func))
	    (when (symbolp name)
	      (get name 'SYS::ARGLIST))))
      (case (car func)
	(LAMBDA (second func))
	(LAMBDA-BLOCK (third func))
	(LAMBDA-CLOSURE (fifth func))
	(LAMBDA-BLOCK-CLOSURE (sixth func)))))

(defun lambda-lexical-variables (func)
  (case (car func)
    (LAMBDA nil)
    (LAMBDA-BLOCK nil)
    ((LAMBDA-CLOSURE LAMBDA-BLOCK-CLOSURE) (second func))))

(defun lambda-parameter-count (func &aux (count 0))
  (dolist (s (lambda-parameter-list func))
    (unless (member s lambda-list-keywords :test #'eq)
	    (incf count)))
  count)

(defun lambda-parameter-name (number parm-list)
  (dolist (s parm-list)
    (unless (and (symbolp s) (char-equal (elt (string s) 0) #\&))
      (decf number))
    (when (minusp number)
      (return (if (listp s) (first s) s)))))

(defun print-restart-options (&optional skip-first-p)
  ;; Give arg of T if this is inside the debugger's CATCH-ABORT.
  (let* ((pcases (if skip-first-p
  		     (cdr (compute-restarts))
  		     (compute-restarts)))
	 (num 0)
	 (lastquit (find 'ABORT pcases :key #'restart-name
			:from-end t))
	 (firstquit (find 'ABORT pcases :key #'restart-name)))
    (unless (null pcases)
      (format t "~%Restart options (Type :C <number>):")
      (dolist (pcase pcases)
        (paranoid-format t "~%~:[~*     ~;:Q~:[, ~;!,~] ~]~d: ~a"
  		(or (eq pcase lastquit) (eq pcase firstquit))
  		(and (eq pcase lastquit) (not (eq pcase firstquit)))
  		num
		pcase)
	(incf num)))))

;;; ----------------------------------------------------------------------
;;; Utilities

(defmacro paranoid-print (arg)
  `(handler-case
    (print ,arg)
    (error (condition) 
	(declare (ignore condition))
	(format t "<<<Error printing object>>>"))))

(defmacro paranoid-format (&rest args)
  `(handler-case 
    (format ,@args)
    (error (condition) 
	(declare (ignore condition))
	(format t "<<<Error printing object>>>"))))
