#ifdef PATHNAME_MSDOS
# Components:
# HOST          NIL
# DEVICE        NIL or :WILD or "A"|...|"Z"
# DIRECTORY     (Startpoint . Subdirs) wobei
#                Startpoint = :RELATIVE | :ABSOLUTE
#                Subdirs = () | (subdir . Subdirs)
#                subdir = :CURRENT (bedeutet ".") or
#                subdir = :PARENT (bedeutet "..") or
#                subdir = :WILD-INFERIORS (bedeutet "...", alle Subdirectories) or
#                subdir = (name . type)
#                 name = :WILD or Simple-String with max. 8 letters
#                 type = :WILD or Simple-String with max. 3 letters
# NAME          NIL or :WILD or Simple-String with max. 8 letters
# TYPE          NIL or :WILD or Simple-String with max. 3 letters
# VERSION       NIL (also :WILD or :NEWEST bei Eingabe)
# Wenn ein Pathname vollstndig spezifiziert sein mu (keine Wildcards),
# ist :WILD, :WILD-INFERIORS nicht erlaubt, bei NAME evtl. also nicht NIL.
# External Notation:       A:\sub1.typ\sub2.typ\name.typ
# with Defaults:             \sub1.typ\sub2.typ\name.typ
# or                                         name.typ
# or                    *:\sub1.typ\*.*\name.*
# or hnliches.
# Statt '\' ist - wie unter DOS blich - also '/' erlaubt.
#endif

#ifdef PATHNAME_UNIX
# Components:
# HOST          NIL
# DEVICE        NIL
# DIRECTORY     (Startpoint . Subdirs) wobei
#                Startpoint = :RELATIVE | :ABSOLUTE
#                Subdirs = () | (subdir . Subdirs)
#                subdir = :WILD-INFERIORS (bedeutet "**" or "...", all Subdirectories) or
#                subdir = Simple-String, evtl. with Wildcard-letters ? and *
# NAME          NIL or
#               Simple-String, evtl. with Wildcard-letters ? and *
#               (also :WILD bei der Eingabe)
# TYPE          NIL or
#               Simple-String, evtl. with Wildcard-letters ? and *
#               (also :WILD bei der Eingabe)
# VERSION       NIL (also :WILD or :NEWEST bei Eingabe)
# A UNIX-Filename wird folgendermaen in Name and Typ aufgespalten:
#   falls kein '.' in Filename: Name = alles, Typ = NIL,
#   falls '.' in Filename: Name = alles vor, Typ = all after the last '.' .
# When a Pathname vollstndig spezifiziert sein mu (keine Wildcards),
# ist :WILD, :WILD-INFERIORS nicht erlaubt, keine Wildcard-letters in den
# Strings, bei NAME evtl. also nicht NIL.
# External Notation:  server:/sub1.typ/sub2.typ/name.typ
# with Defaults:             /sub1.typ/sub2.typ/name.typ
# or                                         name.typ
# or                      /sub1.typ/ ** /sub3.typ/x*.lsp  (without Spaces!)
# or hnliches.
#endif

#ifdef LOGICAL_PATHNAMES
# Components of Logical Pathnames:
# HOST          Simple-String or NIL
# DEVICE        NIL
# DIRECTORY     (Startpoint . Subdirs) wobei
#                Startpoint = :RELATIVE | :ABSOLUTE
#                Subdirs = () | (subdir . Subdirs)
#               subdir = :WILD-INFERIORS (bedeutet "**", all Subdirectories) or
#               subdir = :WILD (bedeutet "*") or
#               subdir = Simple-String, evtl. with Wildcard-letters *
# NAME          NIL or
#               :WILD (bedeutet "*") or
#               Simple-String, evtl. with Wildcard-letters *
# TYPE          NIL or
#               :WILD (bedeutet "*") or
#               Simple-String, evtl. with Wildcard-letters *
# VERSION       NIL or :NEWEST or :WILD or Integer
# Externe Notation: siehe CLtl2 S. 628-629.
#endif

#ifdef LOGICAL_PATHNAMES

local boolean legal_logical_word_char (uintB ch);
local boolean legal_logical_word_char(ch)
  var reg1 uintB ch;
  { ch = up_case(ch);
    if (((ch >= 'A') && (ch <= 'Z'))
        || ((ch >= '0') && (ch <= '9'))
        || (ch == '-')
       )
      return TRUE;
      else
      return FALSE;
  }

#endif


object
parse_namestring(object s, int start, int end, int *ep)
{
	int i, j, k, d;
	object path, *plast = &path, name, type;

	path = Cnil;
	for (i = j = start;  i < end;  ) {
		if (isspace(s->st.st_self[i]))
			break;
#ifdef unix
		if (IS_DIR_SEPARATOR(s->st.st_self[i])) {
#endif
			if (j == 0 && i == 0) {
				i++;
				plast = &CDR(*plast = CONS(Kroot, Cnil));
				j = i;
				continue;
			}
#ifdef unix
			/* BUG FIX by Grant J. Munsey */
			if (i == j) {
				i++;
				j = i;
				continue;
			}
			/* END OF BUG FIX */
			if (i-j == 1 && s->st.st_self[j] == '.') {
				plast = &CDR(*plast = CONS(Kcurrent, Cnil));
			} else if (i-j==2 && s->st.st_self[j]=='.' && s->st.st_self[j+1]=='.') {
				plast = &CDR(*plast = CONS(Kparent, Cnil));
			} else
				plast = &CDR(*plast = CONS(
					make_one(&s->st.st_self[j], i-j), Cnil));
#endif
			i++;
			j = i;
		} else
			i++;
	}
	*ep = i;
	if (i == j) {
		/*  no file and no type  */
		name = Cnil;
		type = Cnil;
		goto L;
	}
	for (k = j, d = -1;  k < i;  k++)
		if (s->st.st_self[k] == '.')
			d = k;
	if (d == -1) {
		/*  no file type  */
#ifdef unix
		if (i-j == 1 && s->st.st_self[j] == '*')
#endif
			name = Kwild;
		else
			name = make_one(&s->st.st_self[j], i-j);
		type = Cnil;
	} else if (d == j) {
		/*  no file name  */
		name = Cnil;
#ifdef unix
		if (i-d-1 == 1 && s->st.st_self[d+1] == '*')
#endif
			type = Kwild;
		else
			type = make_one(&s->st.st_self[d+1], i-d-1);
	} else {
		/*  file name and file type  */
#ifdef unix
		if (d-j == 1 && s->st.st_self[j] == '*')
#endif
			name = Kwild;
		else
			name = make_one(&s->st.st_self[j], d-j);
#ifdef unix
		if (i-d-1 == 1 && s->st.st_self[d+1] == '*')
#endif
			type = Kwild;
		else
			type = make_one(&s->st.st_self[d+1], i-d-1);
	}
L:
	path = tilde_expand(path);
	return(make_pathname(Cnil, Cnil, path, name, type, Cnil));

NO:
	*ep = i;
	return(OBJNULL);
}

@(defun parse_namestring (thing
	&o host
	   (defaults `symbol_value(Vdefault_pathname_defaults)`)
	&k start end junk_allowed
	&a x y)
	int s, e, ee;
@
	check_type_or_pathname_string_symbol_stream(&thing);
	check_type_or_pathname_string_symbol_stream(&defaults);
	defaults = coerce_to_pathname(defaults);
	host = test_optional_host(host);
	if (Null(host)) {
#	  ifdef LOGICAL_PATHNAMES
          if (logpathnamep(defaults))
            { parse_logical = TRUE;
	      host = TheLogpathname(defaults)->pathname_host; }
            else
#         endif
            {
#             if HAS_HOST
              host = ThePathname(defaults)->pathname_host;
#             else
              host = NIL;
#             endif
            }
        }
#       ifdef LOGICAL_PATHNAMES
        elif (logical_host_p(host))
        { parse_logical = TRUE; host = string_upcase(host); }
#       endif

	x = thing;
L:
	switch (type_of(x)) {
	case t_symbol:
	case t_string:
		get_string_start_end(x, start, end, &s, &e);
		for (;  s < e && isspace(x->st.st_self[s]);  s++)
			;
		/* !!!!! Bug Fix. NLG */
		y = parse_namestring(x, s, e - s, &ee);
		if (Null(junk_allowed)) {
			for (;  ee < e - s;  ee++)
				if (!isspace(x->st.st_self[s + ee]))
					break;
			if (y == OBJNULL || ee != e - s)
				FEerror("Cannot parse the namestring ~S~%\
from ~S to ~S.",
					3, x, start, end);
		} else
			if (y == OBJNULL)
				@(return Cnil `MAKE_FIXNUM(s + ee)`)
		start = MAKE_FIXNUM(s + ee);
		break;

	case t_pathname:
		y = x;
		break;

	case t_stream:
		switch ((enum smmode)x->sm.sm_mode) {
		case smm_input:
		case smm_output:
		case smm_probe:
		case smm_io:
			x = x->sm.sm_object1;
			/*
				The file was stored in sm.sm_object1.
				See open.
			*/
			goto L;

		case smm_synonym:
			x = symbol_value(x->sm.sm_object0);
			goto L;

		default:
			goto CANNOT_PARSE;
		}

	default:
	CANNOT_PARSE:
		FEerror("Cannot parse the namestring ~S.", 1, x);
	}
	if (host != Cnil && y->pn.pn_host != Cnil &&
	    host != y->pn.pn_host)
		FEerror("The hosts ~S and ~S do not match.",
			2, host, y->pn.pn_host);
	@(return y start)
@)

LISPFUN(parse_namestring,1,2,norest,key,3,\
        (kw(start),kw(end),kw(junk_allowed)) )
# (PARSE-NAMESTRING thing [host [defaults [:start] [:end] [:junk-allowed]]]),
# CLTL S. 414
  { # Stackaufbau: thing, host, defaults, start, end, junk-allowed.
    var reg6 boolean junk_allowed;
    var reg7 boolean parse_logical = FALSE;
    # Stackaufbau: thing, host, defaults, start, end.
    # 2. Default-Wert fr start ist 0:
    { if (eq(STACK_1,unbound)) { STACK_1 = Fixnum_0; }}
    # 3. host berprfen:
    #if HAS_HOST || defined(LOGICAL_PATHNAMES)
    { var reg2 object host;
      #if HAS_HOST
      host = test_optional_host(STACK_3,FALSE);
      #else
      host = test_optional_host(STACK_3);
      #endif
      if (nullp(host))
        { # host := (PATHNAME-HOST defaults)
          var reg1 object defaults = test_default_pathname(STACK_2);
          #ifdef LOGICAL_PATHNAMES
          if (logpathnamep(defaults))
            { parse_logical = TRUE; host = TheLogpathname(defaults)->pathname_host; }
            else
          #endif
            {
              #if HAS_HOST
              host = ThePathname(defaults)->pathname_host;
              #else
              host = NIL;
              #endif
            }
        }
      #ifdef LOGICAL_PATHNAMES
      elif (logical_host_p(host))
        { parse_logical = TRUE; host = string_upcase(host); }
      #endif
      STACK_3 = host;
    }
    #else
    { test_optional_host(STACK_3); }
    #endif
    # 4. thing must be a String:
    { var reg5 object thing = STACK_4;
      if (xpathnamep(thing)) # Pathname?
        { value1 = thing; # 1. Wert thing
          fertig:
          value2 = STACK_1; mv_count=2; # 2. Wert start
          skipSTACK(5); return;
        }
      if (streamp(thing)) # Stream?
        { thing = as_file_stream(thing);
          value1 = TheStream(thing)->strm_file_name; # 1. Wert: Filename
          goto fertig; # 2. Wert wie oben
        }
      # thing sollte nun wenigstens ein String or Symbol sein:
      if (!stringp(thing))
        { if (!symbolp(thing)) { fehler_thing(thing); }
          thing = Symbol_name(thing); # Symbol -> Symbolname verwenden
          if (!parse_logical)
            {
              #if defined(PATHNAME_UNIX) || defined(PATHNAME_OS2) || defined(PATHNAME_RISCOS)
              # Betriebssystem mit Vorzug fr Kleinbuchstaben
              thing = copy_string(thing); # ja -> mit STRING-DOWNCASE umwandeln
              nstring_downcase(&TheSstring(thing)->data[0],TheSstring(thing)->length);
              #endif
            }
          STACK_4 = thing; # und in den Stack zurckschreiben
        }
      # thing = STACK_4 ist jetzt ein String.
      { # Er wird durchlaufen.
        var zustand z; # laufender Zustand
       {var object string; # String thing
        # Grenzen berprfen, mit thing, start, end als Argumenten:
        pushSTACK(thing); pushSTACK(STACK_(1+1)); pushSTACK(STACK_(0+2));
        test_string_limits(&string,&z.index,&z.count);
        # z.index = Wert des start-Arguments,
        # z.count = Anzahl der Characters.
        z.FNindex = fixnum(z.index);
        # z.FNindex = start-Index als Fixnum.
        string = array_displace_check(string,z.count,&z.index); # Datenvektor holen,
        # z.index = Offset + Startindex = Startoffset
        pushSTACK(string);
       }
        #ifdef LOGICAL_PATHNAMES
        if (parse_logical)
          { pushSTACK(allocate_logpathname());
            # Stackaufbau: ..., Datenvektor, Pathname.
           {var reg1 uintL remaining = parse_logical_pathnamestring(z);
            z.index += z.count-remaining; z.FNindex = fixnum_inc(z.FNindex,z.count-remaining); z.count = remaining;
          }}
          else
        #endif
          {
            pushSTACK(allocate_pathname());
            # Stackaufbau: ..., Datenvektor, Pathname.
            #if HAS_HOST
              # Host-Specification parsen:
              {var reg3 object host;
               { var zustand startz = z; # Start-Zustand
                 var reg1 uintB ch;
                 #if defined(PATHNAME_RISCOS)
                   # Kommt eine Folge von Zeichen, eingeschlossen in '-',
                   # or eine Folge von Zeichen und dann eine ':' ?
                   if (z.count==0) goto no_hostspec; # String schon zu Ende -> kein Host
                   ch = TheSstring(STACK_1)->data[z.index]; # nchstes Character
                   if (ch=='-')
                     { # '-' bergehen:
                       z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                       loop
                         { if (z.count==0) goto no_hostspec; # String schon zu Ende -> kein Host
                           ch = TheSstring(STACK_1)->data[z.index]; # nchstes Character
                           if (ch=='-') break;
                           if (!legal_hostchar(ch)) goto no_hostspec;
                           # gltiges Character bergehen:
                           z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                         }
                       # Host-String bilden:
                       { var reg4 uintL len = z.index - startz.index - 1;
                         if (len==0) goto no_hostspec;
                         host = allocate_string(len);
                         # und fllen:
                        {var reg2 uintB* ptr1 = &TheSstring(STACK_1)->data[startz.index+1];
                         var reg3 uintB* ptr2 = &TheSstring(host)->data[0];
                         dotimesL(len,len, { *ptr2++ = *ptr1++; });
                       }}
                     }
                     else
                     { loop
                         { if (!legal_hostchar(ch)) goto no_hostspec;
                           # gltiges Character bergehen:
                           z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                           if (z.count==0) goto no_hostspec; # String schon zu Ende -> kein Host
                           ch = TheSstring(STACK_1)->data[z.index]; # nchstes Character
                           if (ch==':') break;
                         }
                       # Host-String bilden:
                       { var reg4 uintL len = z.index - startz.index;
                         host = allocate_string(len);
                         # und fllen:
                        {var reg2 uintB* ptr1 = &TheSstring(STACK_1)->data[startz.index];
                         var reg3 uintB* ptr2 = &TheSstring(host)->data[0];
                         dotimesL(len,len, { *ptr2++ = *ptr1++; });
                       }}
                     }
                   # Character '-' bzw. ':' bergehen:
                   z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                   goto hostspec_ok;
                 #else
                   # Kommt eine Folge von alphanumerischen Zeichen und dann ein ':' bzw. '::' ?
                   loop
                     { if (z.count==0) goto no_hostspec; # String schon zu Ende -> kein Host
                       ch = TheSstring(STACK_1)->data[z.index]; # nchstes Character
                       if (!alphanumericp(ch)) break;
                       # alphanumerisches Character bergehen:
                       z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                     }
                   if (!(ch==':')) goto no_hostspec; # kein ':' -> kein Host
                   # Host-String bilden:
                   { var reg4 uintL len = z.index - startz.index;
                     host = allocate_string(len);
                     # und fllen:
                    {var reg2 uintB* ptr1 = &TheSstring(STACK_1)->data[startz.index];
                     var reg3 uintB* ptr2 = &TheSstring(host)->data[0];
                     dotimesL(len,len, { *ptr2++ = *ptr1++; });
                   }}
                   # Character ':' bergehen:
                   z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                   goto hostspec_ok;
                 #endif
                 no_hostspec: # keine Host-Specification
                   z = startz; # zum Start zurck
                   host = STACK_(3+2); # Default-Host
               }
               hostspec_ok:
               # Host eintragen:
               ThePathname(STACK_0)->pathname_host = host;
              }
            #endif
            #if HAS_DEVICE
             #if defined(PATHNAME_ATARI) || defined(PATHNAME_MSDOS) || defined(PATHNAME_OS2)
              # Einbuchstabige Device-Specification und evtl. Seriennummer parsen:
              {var reg3 object device = NIL; # Device := NIL
               #if HAS_SERNR
               var reg4 object seriennummer = NIL; # Seriennummer := NIL
               #endif
               # Drive-Specification parsen:
               # Kommt evtl. ein Buchstabe ('*','A'-'Z','a'-'z'), evtl. eine
               # Seriennummer (Integer >=0, <2^24) und dann ein ':' ?
               { var zustand startz = z; # Start-Zustand
                 var reg1 uintB ch;
                 if (z.count==0) goto no_drivespec; # String schon zu Ende ?
                 ch = TheSstring(STACK_1)->data[z.index]; # nchstes Character
                 ch = up_case(ch); # als Grobuchstabe
                 if (ch == '*')
                   # ch = '*' -> Device := :WILD
                   { device = S(Kwild); }
                 elif ((ch >= 'A') && (ch <= 'Z'))
                   # 'A' <= ch <= 'Z' -> Device := "ch"
                   { var reg1 object string = allocate_string(1); # String der Lnge 1
                     TheSstring(string)->data[0] = ch; # mit ch als einzigem Buchstaben
                     device = string;
                   }
                 else goto no_device;
                 # Device OK, Character bergehen:
                 z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                 if (z.count==0) goto no_drivespec; # String schon zu Ende ?
                 ch = TheSstring(STACK_1)->data[z.index]; # nchstes Character
                 ch = up_case(ch); # als Grobuchstabe
                 no_device:
                 #if HAS_SERNR
                 # Kommt eventuell eine Seriennummer ?
                 if ((ch >= '0') && (ch <= '9'))
                   { # ja.
                     var reg2 uintL akku = 0; # Hilfsregister zum Aufbau der Seriennummer
                     loop
                       { ch = ch - '0'; # Wert der Ziffer
                         akku = 10*akku+ch; # Seriennummer um eine Ziffer erweitern
                         if (akku > (uintL)(bitm(oint_data_len)-1))
                           goto no_drivespec; # >=2^24 geworden -> ist wohl keine Seriennummer
                         # Character bergehen:
                         z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                         if (z.count==0) goto no_drivespec; # String schon zu Ende ?
                         ch = TheSstring(STACK_1)->data[z.index]; # nchstes Character
                         ch = up_case(ch); # als Grobuchstabe
                         # noch eine Ziffer -> zhlt zur Seriennummer
                         if (!((ch >= '0') && (ch <= '9'))) break;
                       }
                     # Seriennummer zu Ende
                     seriennummer = fixnum(akku); # akku als Fixnum
                   }
                 #endif
                 # mit Doppelpunkt abgeschlossen?
                 if (!(ch == ':')) goto no_drivespec;
                 # Character bergehen:
                 z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                 goto drivespec_ok;
                 no_drivespec:
                 # Es ist nicht gelungen, eine Drive-Specification zu parsen.
                 z = startz; # Start-Zustand wiederherstellen
                 device = NIL; # Device := NIL
                 #if HAS_SERNR
                 seriennummer = NIL; # Seriennummer := NIL
                 #endif
               }
               drivespec_ok:
               ThePathname(STACK_0)->pathname_device = device; # Device eintragen
               #if HAS_SERNR
               if (!NIL_IS_CONSTANT) { pushSTACK(seriennummer); } # Seriennummer retten (kann NIL or ein Fixnum sein)
               { var reg1 object new_cons = allocate_cons(); # neues Cons
                 if (!NIL_IS_CONSTANT) { seriennummer = popSTACK(); } # Seriennummer zurck
                 Car(new_cons) = seriennummer; # = (cons Seriennummer NIL)
                 ThePathname(STACK_0)->pathname_directory = new_cons;
                 pushSTACK(new_cons);
                 new_cons = allocate_cons(); # neues Cons fr Startpoint
                 Cdr(STACK_0) = new_cons; # verlngert (pathname-directory Pathname)
                 STACK_0 = new_cons; # neues (last (pathname-directory Pathname))
               }
               #define HAVE_Startpoint_Cons
               #endif
              }
             #endif
            #ifndef HAVE_Startpoint_Cons # falls nicht oben schon erledigt
            # Directory-Start eintragen:
            { var reg1 object new_cons = allocate_cons(); # neues Cons fr Startpoint
              ThePathname(STACK_0)->pathname_directory = new_cons;
              pushSTACK(new_cons); # neues (last (pathname-directory Pathname))
            }
            #endif
            # Stackaufbau: ..., Datenvektor, Pathname, (last (pathname-directory Pathname)).
            # Subdirectories parsen:
            # Trennzeichen zwischen subdirs ist unter MSDOS sowohl '\' als auch '/':
            #if defined(PATHNAME_MSDOS) || defined(PATHNAME_OS2)
             #define slashp(c)  (((c) == '\\') || ((c) == '/'))
	    #else
             #define slashp(c)  ((c) == '/')
            #endif
            {
              #if defined(USER_HOMEDIR) && defined(PATHNAME_UNIX)
              # Falls sofort ein '~' kommt, wird bis zum nchsten '/' or Stringende
              # ein Username gelesen und das Home-Directory dieses Users eingesetzt:
              if ((!(z.count == 0)) && (TheSstring(STACK_2)->data[z.index] == '~'))
                # Es kommt sofort ein '~'.
                { # Character bergehen:
                  z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                 {var reg6 object userhomedir; # Pathname des User-Homedir
                  # nchsten '/' suchen:
                  var reg2 uintB* charptr = &TheSstring(STACK_2)->data[z.index];
                  var reg3 uintL charcount = 0;
                  { var reg4 uintL count;
                    dotimesL(count,z.count,
                      { if (*charptr++ == '/') break;
                        charcount++;
                      });
                  }
                  # Username hat charcount Zeichen
                  if (charcount==0)
                    { userhomedir = O(user_homedir); } # nur '~' -> User-Homedir
                    else
                    { # Username als ASCIZ-String bauen:
                      var reg5 object username = allocate_string(charcount+1);
                      { var reg1 uintB* charptr2 = &TheSstring(username)->data[0];
                        var reg4 uintL count;
                        charptr = &TheSstring(STACK_2)->data[z.index];
                        dotimespL(count,charcount, { *charptr2++ = *charptr++; } );
                        *charptr2 = '\0';
                      }
                      # Dessen Home-Directory aus dem Passwort-File holen:
                      begin_system_call();
                      errno = 0;
                     {var reg1 struct passwd * userpasswd = getpwnam(TheAsciz(username));
                      end_system_call();
                      if (userpasswd == (struct passwd *)NULL) # erfolglos?
                        { if (!(errno==0)) { OS_error(); } # Error melden
                          # sonst: Fehler
                          pushSTACK(username);
                          pushSTACK(S(parse_namestring));
                          fehler(error,
                                 DEUTSCH ? "~: Es gibt keinen Benutzer mit Namen ~." :
                                 ENGLISH ? "~: there is no user named ~" :
                                 FRANCAIS ? "~ : Il n'y a pas d'utilisateur de nom ~." :
                                 ""
                                );
                        }
                      userhomedir = asciz_dir_to_pathname(userpasswd->pw_dir); # Homedir als Pathname
                    }}
                  # Directory aus dem Pathname userhomedir kopieren:
                  # (copy-list dir) = (nreconc (reverse dir) nil),
                  # dabei dessen letztes Cons merken.
                  userhomedir = reverse(ThePathname(userhomedir)->pathname_directory);
                  STACK_0 = userhomedir; userhomedir = nreconc(userhomedir,NIL);
                  ThePathname(STACK_1)->pathname_directory = userhomedir;
                  # username-Characters bergehen:
                  z.index += charcount; z.FNindex = fixnum_inc(z.FNindex,charcount); z.count -= charcount;
                  # Falls der String zu Ende ist: fertig,
                  # sonst kommt sofort ein '/', es wird bergangen:
                  if (z.count==0)
                    { pushSTACK(NIL); pushSTACK(NIL); goto after_name_type; } # Name und Typ := NIL
                  # Character bergehen:
                  z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                }}
              else
              #endif
              #if defined(PATHNAME_UNIX) && 0 # Wozu braucht man das, auer fr $HOME ?
              # Falls sofort ein '$' kommt, wird bis zum nchsten '/' or Stringende
              # eine Environment-Variable gelesen und ihr Wert eingesetzt:
              if ((!(z.count == 0)) && (TheSstring(STACK_2)->data[z.index] == '$'))
                # Es kommt sofort ein '$'.
                { # Character bergehen:
                  z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                 {var reg6 object envval_dir;
                  # nchsten '/' suchen:
                  var reg2 uintB* charptr = &TheSstring(STACK_2)->data[z.index];
                  var reg3 uintL charcount = 0;
                  { var reg4 uintL count;
                    dotimesL(count,z.count,
                      { if (*charptr++ == '/') break;
                        charcount++;
                      });
                  }
                  # Environment-Variable hat charcount Zeichen,
                  # als ASCIZ-String bauen:
                  { var reg5 object envvar = allocate_string(charcount+1);
                    { var reg1 uintB* charptr2 = &TheSstring(envvar)->data[0];
                      var reg4 uintL count;
                      charptr = &TheSstring(STACK_2)->data[z.index];
                      dotimesL(count,charcount, { *charptr2++ = *charptr++; } );
                      *charptr2 = '\0';
                    }
                    # Dessen Wert holen:
                    begin_system_call();
                   {var reg1 const char* envval = getenv(TheAsciz(envvar));
                    end_system_call();
                    if (envval==NULL)
                      { pushSTACK(envvar);
                        pushSTACK(S(parse_namestring));
                        fehler(error,
                               DEUTSCH ? "~: Es gibt keine Environment-Variable ~." :
                               ENGLISH ? "~: there is no environment variable ~" :
                               FRANCAIS ? "~ : Il n'y a pas de variable ~ dans l'environnement." :
                               ""
                              );
                      }
                    envval_dir = asciz_dir_to_pathname(envval); # Wert der Variablen als Pathname
                  }}
                  # Directory aus dem Pathname envval_dir kopieren:
                  # (copy-list dir) = (nreconc (reverse dir) nil),
                  # dabei dessen letztes Cons merken.
                  envval_dir = reverse(ThePathname(envval_dir)->pathname_directory);
                  STACK_0 = envval_dir; envval_dir = nreconc(envval_dir,NIL);
                  ThePathname(STACK_1)->pathname_directory = envval_dir;
                  # envvar-Characters bergehen:
                  z.index += charcount; z.FNindex = fixnum_inc(z.FNindex,charcount); z.count -= charcount;
                  # Falls der String zu Ende ist: fertig,
                  # sonst kommt sofort ein '/', es wird bergangen:
                  if (z.count==0)
                    { pushSTACK(NIL); pushSTACK(NIL); goto after_name_type; } # Name und Typ := NIL
                  # Character bergehen:
                  z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                }}
              else
              #endif
              #if defined(PATHNAME_UNIX) || defined(PATHNAME_ATARI) || defined(PATHNAME_MSDOS) || defined(PATHNAME_OS2)
              # Falls sofort ein '\' bzw. '/' kommt, wird er bergangen, und es kommt
              # :ABSOLUTE (sonst :RELATIVE) als erstes subdir:
              if ((!(z.count == 0)) && slashp(TheSstring(STACK_2)->data[z.index]))
                # Es kommt sofort ein '\' bzw. '/'.
                { # Character bergehen:
                  z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                  Car(STACK_0) = S(Kabsolute); # Startpoint = :ABSOLUTE
                }
                else
                # Es kommt nicht sofort ein '\' bzw. '/'.
                { Car(STACK_0) = S(Krelative); } # Startpoint = :RELATIVE
              #endif
              loop
                { # Versuche, ein weiteres Unterdirectory zu parsen.
                  #ifdef PATHNAME_EXT83
                    # Kommt '.\' or '..\' or '...\' ?
                    if ((!(z.count == 0)) && (TheSstring(STACK_2)->data[z.index] == '.'))
                      { # nchstes Character ist ein '.'.
                        var zustand subdirz = z; # Zustand beim Start des Subdirectories
                        # Character bergehen:
                        z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                        if (z.count == 0) goto no_dots; # String schon zu Ende ?
                       {var reg1 uintB ch = TheSstring(STACK_2)->data[z.index]; # nchstes Character
                        if (slashp(ch))
                          # '.\' angetroffen -> (cons :CURRENT NIL) bauen
                          { pushSTACK(S(Kcurrent)); goto dots; }
                        if (!(ch == '.')) goto no_dots;
                        # zweites Character war auch ein '.'.
                        # Character bergehen:
                        z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                        if (z.count == 0) goto no_dots; # String schon zu Ende ?
                        ch = TheSstring(STACK_2)->data[z.index]; # nchstes Character
                        if (slashp(ch))
                          # '..\' angetroffen -> (cons :PARENT NIL) bauen
                          { pushSTACK(S(Kparent)); goto dots; }
                        if (!(ch == '.')) goto no_dots;
                        # drittes Character war auch ein '.'.
                        # Character bergehen:
                        z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                        if (z.count == 0) goto no_dots; # String schon zu Ende ?
                        ch = TheSstring(STACK_2)->data[z.index]; # nchstes Character
                        if (slashp(ch))
                          # '...\' angetroffen -> (cons :WILD-INFERIORS NIL) bauen
                          { pushSTACK(S(Kwild_inferiors)); goto dots; }
                        goto no_dots;
                       }
                        dots:
                        # '.\' or '..\' or '...\' angetroffen, Keyword im Stack.
                        # Character '\' bergehen:
                        z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                        goto subdir_ok;
                        no_dots:
                        z = subdirz; # Zustand wiederherstellen
                      }
                    # Versuche, normale 'name.typ'-Syntax zu parsen:
                    pushSTACK(NIL); # dummy
                    { # Name, hat max. 8 Buchstaben:
                      var reg1 object name = parse_name_or_type(&z,8,NIL);
                      STACK_0 = name;
                    }
                    # Versuche, '.typ'-Syntax zu parsen:
                    { var reg1 object type;
                      if ((!(z.count==0)) && (TheSstring(STACK_3)->data[z.index] == '.'))
                        { # Es kommt ein '.'. Character bergehen:
                          z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                          # Typ, hat max. 3 Buchstaben:
                          type = parse_name_or_type(&z,3,O(leer_string));
                        }
                        else
                        { type = NIL; }
                      pushSTACK(type);
                    }
                    # Stackaufbau: ...,
                    #   Datenvektor, Pathname, (last (pathname-directory Pathname)),
                    #   name, type.
                    # Kommt sofort ein '\', so war es ein Unterdirectory,
                    # sonst ist der Pathname beendet:
                    if ((z.count==0) || !slashp(TheSstring(STACK_4)->data[z.index])) break;
                    # Es kommt ein '\'. Character bergehen:
                    z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                    # name=NIL -> durch "" ersetzen:
                    if (eq(STACK_1,NIL)) { STACK_1 = O(leer_string); }
                    # type=NIL -> durch "" ersetzen:
                    if (eq(STACK_0,NIL)) { STACK_0 = O(leer_string); }
                    { var reg1 object new_cons = allocate_cons(); # neues Cons
                      Cdr(new_cons) = popSTACK(); # type
                      Car(new_cons) = popSTACK(); # name
                      # new_cons = (cons name type)
                      pushSTACK(new_cons);
                    }
                    subdir_ok:
                  #endif
                  #ifdef PATHNAME_NOEXT
                    { var reg3 uintL z_start_index = z.index; # Index beim Start
                      loop
                        { var reg2 uintB ch;
                          if (z.count == 0) break;
                          ch = TheSstring(STACK_2)->data[z.index]; # nchstes Character
                          if (!legal_namechar(ch)) break; # gltiges Character ?
                          # ja -> Teil des Namens
                          # Character bergehen:
                          z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                        }
                      # Ende des Namens erreicht.
                      # Name := Teilstring von STACK_2 von z_start_index (einschlielich)
                      #                                bis z.index (ausschlielich).
                     {var reg3 uintL len = z.index - z_start_index;
                      var reg4 object string = allocate_string(len); # String der Lnge len
                      # fllen:
                      var reg1 uintB* ptr1 = &TheSstring(STACK_2)->data[z_start_index];
                      var reg2 uintB* ptr2 = &TheSstring(string)->data[0];
                      dotimesL(len,len, { *ptr2++ = *ptr1++; });
                      # Name fertig.
                      pushSTACK(string);
                    }}
                    # Kommt sofort ein '/' bzw. '\', so war es ein Unterdirectory,
                    # sonst ist der Pathname beendet:
                    if ((z.count==0) || !slashp(TheSstring(STACK_3)->data[z.index]))
                      # Nein -> war der Name und kein Subdir.
                      break;
                    # Es kommt ein '/' bzw. '\'. Character bergehen:
                    z.index++; z.FNindex = fixnum_inc(z.FNindex,1); z.count--;
                    # Stackaufbau: ...,
                    #   Datenvektor, Pathname, (last (pathname-directory Pathname)),
                    #   subdir.
                    # War es '**' or '...' ?
                    if (equal(STACK_0,O(wildwild_string)) || equal(STACK_0,O(punktpunktpunkt_string)))
                      { STACK_0 = S(Kwild_inferiors); } # ja -> durch :WILD-INFERIORS ersetzen
                  #endif
                  # (pathname-directory pathname) um Subdir STACK_0 verlngern:
                  { var reg1 object new_cons = allocate_cons(); # neues Cons
                    Car(new_cons) = popSTACK(); # = (cons subdir NIL)
                    Cdr(STACK_0) = new_cons; # verlngert (pathname-directory Pathname)
                    STACK_0 = new_cons; # neues (last (pathname-directory Pathname))
                  }
                }
              #if defined(PATHNAME_EXT83)
              # Stackaufbau: ...,
              #   Datenvektor, Pathname, (last (pathname-directory Pathname)),
              #   name, type.
              # Name und Typ in Pathname eintragen:
              { var reg3 object type = popSTACK();
                var reg2 object name = popSTACK();
                skipSTACK(1); # Directory ist schon eingetragen
               {var reg1 object pathname = STACK_0;
                ThePathname(pathname)->pathname_name = name;
                ThePathname(pathname)->pathname_type = type;
              }}
              #endif
              #ifdef PATHNAME_NOEXT
              # Stackaufbau: ..., Datenvektor, Pathname, (last (pathname-directory Pathname)),
              #              string.
              split_name_type(0); # String STACK_0 in Name und Typ aufspalten
              after_name_type:
              # Stackaufbau: ..., Datenvektor, Pathname, (last (pathname-directory Pathname)),
              #              name, type.
              # Name und Typ in Pathname eintragen:
              { var reg3 object type = popSTACK();
                var reg2 object name = popSTACK();
                skipSTACK(1); # Directory ist schon eingetragen
                # name="" durch Name=NIL ersetzen:
                if (equal(name,O(leer_string))) { name = NIL; }
               {var reg1 object pathname = STACK_0;
                ThePathname(pathname)->pathname_name = name;
                ThePathname(pathname)->pathname_type = type;
              }}
              #endif
            }
            #undef slashp
          }
        # Pathname fertig.
        # Stackaufbau: ..., Datenvektor, Pathname.
        if (!junk_allowed)
          # berprfen, ob keine Zeichen mehr brig sind:
          if (!(z.count == 0))
            { pushSTACK(z.FNindex); # letzter Index
              pushSTACK(STACK_(4+2+1)); # thing
              pushSTACK(S(parse_namestring));
              fehler(error,
                     ENGLISH ? "~: syntax error in filename ~ at position ~" :
                     ""
                    );
            }
        #ifdef LOGICAL_PATHNAMES
        if (parse_logical)
          { if (!nullp(STACK_(3+2)))
              # Hosts mssen bereinstimmen, vgl. CLtL2 S. 629
              if (!equal(STACK_(3+2),TheLogpathname(STACK_0)->pathname_host))
                { pushSTACK(STACK_0);
                  pushSTACK(TheLogpathname(STACK_(0+1))->pathname_host);
                  pushSTACK(STACK_(3+2+2));
                  pushSTACK(S(parse_namestring));
                  fehler(error,
                         ENGLISH ? "~: hosts ~ and ~ of ~ should coincide" :
                         ""
                        );
          }     }
        #endif
        value1 = STACK_0; # Pathname als 1. Wert
        value2 = z.FNindex; # Index als 2. Wert
        mv_count=2; # 2 Werte
        skipSTACK(5+2); return;
  } } }


#ifdef LOGICAL_PATHNAMES
# define logpathnamep(obj)  (type_of(obj) == t_logpathname)
#endif

Lpathname_host(int narg, object pname)
{
	check_arg(1);

	check_type_or_pathname_string_symbol_stream(&pname);
	pname = coerce_to_pathname(pname);
#	ifdef LOGICAL_PATHNAMES
	if (logpathnamep(pname))
	 VALUES(0) = TheLogpathname(pname)->pn.pn_host;
	else
#	endif
	 VALUES(0) = pname->pn.pn_host;
	RETURN(1);
}
