;;; Soluzione Sherlis:
;;; 1. call_or_link aggiunge a Vlink_array coppie instr.addr/previous.cont
;;;    (e' piu' sicuro che metterle nella plist del symbol)
;;; 2. clear_compiler_properties (da defun) chiama use_fast_links per
;;;    ripristinare. Lo fa scorrendo interamente Vlink_array.
;;;
;;; Sherlis inizializza il codice compilato con la funzione do_init, che
;;; interpreta le forme di inizializzazione. (alternativa a nostra init_label)
;;; Usa codice per la lettura di .data
;;;
;;; Il comando :bl da' i locali di tutte le funzioni (anche compilate)
;;; prendendoli dal VS.

(in-package 'si)

(setq si:*function-entry-table*
  (make-array 64 :fill-pointer 0 :adjustable t :element-type 'fixnum))

;;; this is not correct since entry are not always sorted.
(defun record-function-entry (sym entry)
  (let* ((next (fill-pointer *entry-array*))
	 (last-entry (when (> next 0)
		       (aref *entry-array* (1- next)))))
    (vector-push-extend (cons sym entry) *entry-array*)
    (when (and last-entry
	       (< entry (cdr last-entry)))
      (sort *entry-array* #'< :key #'cdr))))

(defun find-function-entry (pc)
  ;; binary search
  (do* ((low 0)
	(high (fill-pointer *entry-array*))
	(mid) (entry)
	(sym) (addr))
       (())
    (setq mid (floor (+ low high) 2)
	  entry (aref *entry-array* mid)
	  sym (car entry)
	  addr (cdr entry))
    (when (>= (1+ low) high)
      (let* ((n (1+ low))
	     (entry1 (aref *entry-array* n))
	     (sym1 (car entry1))
	     (addr1 (cdr entry1)))
	(return
	  (if sym1
	      (if (<= addr1 pc)
		  sym1
		  (when (and sym (<= addr pc)) sym))
	      (when (and sym (<= addr pc)) sym)))))
    (if sym
	(cond ((< pc addr)
	       (setq high mid))
	      ((> pc addr)
	       (setq low mid))
	      (t (return addr)))
	(setq high mid))))

(defun show-entries ()
  (dotimes (i (fill-pointer *entry-array*))
    (let ((entry (aref *entry-array* i)))
    (format t "~%~A: ~D" (car entry) (cdr entry)))))

(defun backtrace (&optional (stream t))
  (do* ((fp 0)
	(pc 0)
	(stack-base 0)
	(fun))
       ((> fp stack-base))
    (multiple-value-setq (fp pc stack-base)
      (get-previous-frame fp))
    (when (setq fun (find-function-entry pc))
      (format stream "> ~A" fun))))

(Clines "
/* #include <setjmp.h> */

#define FP_INDEX 12
#define PC_INDEX 14
")

(defunc get-previous-frame (p)
"{ int *fp;
  int buf[16];
  if ((fp = fix(p)) == 0) {
     setjmp(buf);
     fp = buf[FP_INDEX];}
  else
     fp = vs_base[0]->big.big_car;
  vs_top += 2;
  vs_base[0] = bignum1(fp[3]);		/* previous FP	*/
  vs_base[1] = MAKE_FIXNUM(fp[4]);	/* previous PC	*/
  vs_base[2] = bignum1(clwp->lwp_cs_org);/* stack base */
}")

#|
(defun b (n &optional (stream t))
  (do* ((fp 0)
	(pc 0)
	(stack-base 0)
	(fun))
       ((= n 0))
    (multiple-value-setq (fp pc stack-base)
      (get-previous-frame fp))
    (when (setq fun (find-function-entry pc))
      (format stream "> ~A" fun))
    (decf n)))
|#
