;; ---------- stack-let for consing on stack ---------

(in-package 'compiler)
(export 'stack-let)

;; Usage:  (stack-let ((a (cons 1 2)) (b (cons 3 4))) (foo a) (print b) 7)
;;  where foo must not keep a copy of `a', since the cons will be formed
;;  on the c stack.

(setf (get 'stack-let 'c1special) 'c1stack-let)

(defmacro stack-let (&rest x) (cons `let x))

(defun c1stack-let (args &aux npairs nums)
  (let ((pairs (car args))
	)
    (dolist (v pairs)
	    (push
	     (cond ((atom v) v)
		   ((let ((var (car v))
			  (val (second v)))
		      (and (consp val)
			   (or (eq (car val) 'cons)
			       (and (eq (car val) 'list)
				    (null (cddr val))
				    (setq val `(cons ,(second val) nil))))
			   (progn
			     (push (next-cvar) nums)
			     `(,var (stack-cons ,(car nums) ,@ (cdr val)))))))
		   (t (cmpwarn "Stack let = regular let for ~a ~a"
			       v (cdr args))
		      v))
	     npairs))
    (let ((result (c1expr (cons 'let (cons (nreverse npairs) (cdr args))))))
      (list 'stack-let (second result) nums result))))

(setf (get 'stack-let 'c2) 'c2stack-let)

(defun c2stack-let (nums form)
  (let ((n (next-cvar)))
    (wt-nl "{Cons_Macro" n ";")
    (c2expr form)
    (wt "}")
    (wt-h
     "#define Cons_Macro" n (format nil " struct cons ~{STcons~a ~^,~};" nums)
     )))

(push '((fixnum t t) t nil nil
	"(STcons#0.t=t_cons,STcons#0.m=0,STcons#0.c_car=(#1),
              STcons#0.c_cdr=(#2),(object)&STcons#0)")
        (get 'stack-cons 'inline-always))    

;;;; new version:
(defun c1stack-let (args &aux npairs nums)
  (labels ((expand-cons (var val)
	   (let (car-var cdr-var)
	     (when (and (consp val)
			(or (eq 'cons (car val))
			    (and (eq (car val) 'list)
				 (if (null (cddr val))
				     (setq val `(cons ,(second val) nil))
				     (setq val `(cons ,(second val) (list ,@(cddr val))))))))
	       (when (is-cons (second val))
		 (setf (second val) (list-to-cons (second val)))
		 (push (setq car-var (gentemp)) nums)
		 (expand-cons car-var (second val)))
	       (when (is-cons (third val))
		 (setf (third val) (list-to-cons (third val)))
		 (push (setq cdr-var (gentemp)) nums)
		 (expand-cons cdr-var (third val)))
	       (push (next-cvar) nums)
	       (push `(,var (stack-cons 
			     ,@(if car-var 
				   (list (car nums) car-var) 
				   (if cdr-var 
				       (list (car nums) (cadr val))
				       (list (car nums))))
			     ,@(if cdr-var (list cdr-var) (cdr val))))
		     npairs)
	       (print npairs)))))
    (dolist (v (car args))
      (cond ((atom v) (push v npairs))
	    ((expand-cons (first v) (second v)))
	    (t (cmpwarn "Stack let = regular let for ~a ~a"
			v (cdr args))
	       (push v npairs))))
    (let ((result (c1expr (cons 'let* (cons (nreverse npairs) (cdr args))))))
      (list 'stack-let (second result) nums result))))

(defun is-cons (x)
  (and (listp x) 
        (or (eq 'cons (car x))
	    (eq 'list (car x)))))

(defun list-to-cons (x)
  (if (eq 'cons (car x))
      x
      `(cons 
	,(second x)
	,(if (null (cddr x))
	     nil
	     `(list ,@(cddr x))))))


;; (defun foo (x) (stack-let ((y (cons x (cons x x)))) (foo y)))
;; (LET ((g0 (stack-cons 1 x x))) (LET ((Y (STACK-CONS 2 x g0))) (FOO Y)))
