/*
    cfun.c -- Compiled functions.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.

    ECoLisp is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    See file '../Copyright' for full details.
*/


#include "config.h"

#ifdef PDE
extern object Sdefun, Sdefmacro;
#endif PDE

object
make_cfun(int (*self)(), object name, struct codeblock *cblock)
{
	object cf;

	cf = alloc_object(t_cfun);
	cf->cf.cf_self = self;
	cf->cf.cf_name = name;
	cf->cf.cf_block = cblock;
	return(cf);
}

object
make_cclosure(int (*self)(), object env, struct codeblock *block)
{
	object cc;

	cc = alloc_object(t_cclosure);
	cc->cc.cc_self = self;
	cc->cc.cc_env = env;
	cc->cc.cc_block = block;
	return(cc);
}

object
MF(object sym, int (*self)(), struct codeblock *block)
{
	object cf;

	if (type_of(sym) != t_symbol)
		not_a_symbol(sym);
	if (SPECIAL(sym) && sym->s.s_mflag)
		sym->s.s_sfdef = NOT_SPECIAL;
	clear_compiler_properties(sym);
#ifndef RUNTIME
	record_fun_entry(sym, self);
#endif
#ifdef PDE
	record_source_pathname(sym, Sdefun);
#endif PDE
	cf = alloc_object(t_cfun);
	cf->cf.cf_self = self;
	cf->cf.cf_name = sym;
	cf->cf.cf_block = block;
	SYM_FUN(sym) = cf;
	sym->s.s_mflag = FALSE;
}

/*
 *----------------------------------------------------------------------
 *
 *  MFkey --
 *	in the keyword table of the function, replaces indexes into the VV
 *	vector with the actual symbols present there.
 *
 *  Returns:
 *	the cfun object
 *
 *----------------------------------------------------------------------
 */
init_keywords(int nkey, intUobject *keys, struct codeblock *block)
{ int i; object *VV = block->cd_data->v.v_self;
  for (i = 0; i < nkey; i++)
    keys[i].o = VV[keys[i].i];
}

object
MFkey(object sym, int (*self)(), struct codeblock *block,
      int nkey, intUobject *keys)
{ init_keywords(nkey, keys, block);
  return(MF(sym, self, block));
}

object
MM(object sym, int (*self)(), struct codeblock *block)
{
	object cf;

	if (type_of(sym) != t_symbol)
		not_a_symbol(sym);
	if (SPECIAL(sym) && sym->s.s_mflag)
		sym->s.s_sfdef = NOT_SPECIAL;
	clear_compiler_properties(sym);
#ifndef RUNTIME
	record_fun_entry(sym, self);
#endif
#ifdef PDE
	record_source_pathname(sym, Sdefmacro);
#endif PDE
	cf = alloc_object(t_cfun);
	cf->cf.cf_self = self;
	cf->cf.cf_name = sym;
	cf->cf.cf_block = block;
	SYM_FUN(sym) = cf;
	sym->s.s_mflag = TRUE;
}

object
make_function(char *s, int (*f)())
{
	object x;

	x = make_ordinary(s);
	SYM_FUN(x) = make_cfun(f, x, NULL);
	x->s.s_mflag = FALSE;
#ifndef RUNTIME
	record_fun_entry(x, f);
#endif
	return(x);
}

object
make_si_function(char *s, int (*f)())
{
	object x;

	x = make_si_ordinary(s);
	SYM_FUN(x) = make_cfun(f, x, NULL);
	x->s.s_mflag = FALSE;
#ifndef RUNTIME
	record_fun_entry(x, f);
#endif
	return(x);
}

object
make_special_form(char *s, int (*f)())
{
	object x;
	x = make_ordinary(s);
	x->s.s_sfdef = f;
#ifndef RUNTIME
	record_fun_entry(x, f);
#endif
	return(x);
}

siLcompiled_function_name(int narg, object fun)
{
	check_arg(1);

	if (type_of(fun) == t_cfun)
		VALUES(0) = fun->cf.cf_name;
	else if (type_of(fun) == t_cclosure)
		VALUES(0) = Cnil;
	else
		FEerror("~S is not a compiled-function.", 1, fun);
	RETURN(1);
}


#ifndef RUNTIME

#define FUN_TABLE_INC 256
extern caddr_t	*function_entry_table;
extern int	function_entries_max;
extern int	function_entries;

/*----------------------------------------------------------------------
 * fun_entry_search --
 *	function_entry_table is an array containing alternated addr, sym values
 *	sorted in increasing addr value.
 * Result:
 *	  the index of the largest addr which is smaller than key
 *	  -2 if no such addr is present
 *----------------------------------------------------------------------
 */
int
fun_entry_search(caddr_t key)
{
  caddr_t *table = function_entry_table;
  int len = function_entries;
  int low = 0;
  int high = len;
  int mid, probe;
  caddr_t entry;
  if (len == 0)
    return(-2);
  while (TRUE) {
    mid = (low + high) / 2;
    probe = mid * 2;
    entry = table[probe];
    if (entry == key)
      return(probe);
    if (entry < key) {
      if (mid + 1 == len || table[probe+2] > key)
	return(probe);
      else
	low = mid;
    } else {
      if (probe == 0)
	return(-2);
      else
	high = mid;
    }
  }
}

/*
 *----------------------------------------------------------------------
 * record_fun_entry --
 *	records the code start of function bound to symbol, so that
 *	one can determine which function is executing
 *
 *----------------------------------------------------------------------
 */
record_fun_entry(object sym, caddr_t addr)
{
  object def;
  register int i, end;

  end = 2*function_entries;
  def = SYM_FUN(sym);
  if (def != OBJNULL && type_of(def) == t_cfun) {
    /* clear previous definition */
    caddr_t prevaddr = (caddr_t)def->cf.cf_self;
    i = fun_entry_search(prevaddr);
    if (i >= 0 && function_entry_table[i] == prevaddr) {
      function_entries--;
      end -= 2;
      memmove(&function_entry_table[i], &function_entry_table[i+2],
	     sizeof(caddr_t) * (end - i));
    }
  }
  i = fun_entry_search(addr);
  if (i < 0 || function_entry_table[i] != addr) {
    if (2*function_entries_max == end) {
      function_entries_max += FUN_TABLE_INC;
      function_entry_table =
	(caddr_t *)realloc(function_entry_table,
			   2 * function_entries_max * sizeof(caddr_t));
    }
    i += 2;
    memmove(&function_entry_table[i+2], &function_entry_table[i],
	   sizeof(caddr_t) * (end - i));
    function_entries++;
  }
  function_entry_table[i++] = addr;
  function_entry_table[i++] = (caddr_t)sym;
}

object
get_function_entry(caddr_t addr)
{
  int i;
  i = fun_entry_search(addr);
  if (i >= 0)
    return((object)function_entry_table[i+1]);
  else
    return(OBJNULL);
}

#endif RUNTIME

init_cfun()
{
	make_si_function("COMPILED-FUNCTION-NAME", siLcompiled_function_name);
}
