/*
    cmpaux.c -- Auxiliaries used in compiled Lisp code.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.

    ECoLisp is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    See file '../Copyright' for full details.
*/


#include "config.h"

object
make_list(int i)
{
  object x = Cnil;
  while (i-- > 0)
    x = CONS(Cnil, x);
  return x;
}

siLspecialp(int narg, object sym)
{
	check_arg(1);
	VALUES(0) = (type_of(sym) == t_symbol &&
		     (enum stype)sym->s.s_stype == stp_special) ? Ct : Cnil;
	RETURN(1);
}

int
ifloor(int x, int y)
{
	if (y == 0)
		FEerror("Zero divizor", 0);
	else if (y > 0)
		if (x >= 0)
			return(x/y);
		else
			return(-((-x+y-1))/y);
	else
		if (x >= 0)
			return(-((x-y-1)/(-y)));
		else
			return((-x)/(-y));
}

int
imod(int x, int y)
{
	return(x - ifloor(x, y)*y);
}

/*
 * ----------------------------------------------------------------------
 *	Conversions to C
 * ----------------------------------------------------------------------
 */

char
object_to_char(object x)
{
	if (FIXNUMP(x)) return(fix(x));	/* Immediate fixnum */

	switch (type_of(x)) {
	  /*	case t_fixnum: return fix(x);	*/
	case t_bignum:
		return x->big.big_limbs[0];
	case t_character:
		return char_code(x);
	default:
		FEerror("~S cannot be coerced to a C char.", 1, x);
	}
}

int
object_to_int(object x)
{
	if (FIXNUMP(x)) return(fix(x));	/* Immediate fixnum */

	switch (type_of(x)) {
	  /*	case t_fixnum: return fix(x);	*/
	case t_character:
		return char_code(x);
	case t_bignum:
		return x->big.big_limbs[0] * big_sign(x);
	case t_ratio:
		return number_to_double(x);
	case t_shortfloat:
		return sf(x);
	case t_longfloat:
		return lf(x);
	default:
		FEerror("~S cannot be coerced to a C int.", 1, x);
	}
}

char *
object_to_string(object x)
{
  extern VOID *malloc();
  switch (type_of(x)) {
  case t_string:
  case t_symbol:
    return(x->st.st_self);
  case t_fixnum: {
    char *num = malloc(12);
    sprintf(num, "%d", fix(x));
    return(num);
  }
  case t_character: {
    char *c = malloc(2);
    c[0] = char_code(x);
    c[1] = '\0';
    return c;
  }
  case t_pathname:
    return namestring(x)->st.st_self;
  default:
    FEerror("~S cannot be coerced to a C string.", 1, x);
  }
}

float
object_to_float(object x)
{
	if (FIXNUMP(x)) return(fix(x));	/* Immediate fixnum */

	switch (type_of(x)) {
	  /*	case t_fixnum: return fix(x);	*/
	case t_character:
		return char_code(x);
	case t_bignum:
	case t_ratio:
		return number_to_double(x);
	case t_shortfloat:
		return sf(x);
	case t_longfloat:
		return lf(x);
	default:
		FEerror("~S cannot be coerced to a C float.", 1, x);
	}
}

double
object_to_double(object x)
{

	if (FIXNUMP(x)) return(fix(x));	/* Immediate fixnum */

	switch (type_of(x)) {
	  /*	case t_fixnum: return fix(x);	*/
	case t_character:
		return char_code(x);
	case t_bignum:
	case t_ratio:
		return number_to_double(x);
	case t_shortfloat:
		return sf(x);
	case t_longfloat:
		return lf(x);
	default:
		FEerror("~S cannot be coerced to a C double.", 1, x);
	}
}

aref_bv(object x, int index)
{
  index += x->bv.bv_offset;
  return ((x->bv.bv_self[index/CHAR_BIT] & (0200>>index%CHAR_BIT)) != 0);
}

aset_bv(object x, int index, int value)
{
  index += x->bv.bv_offset;
  if (value == 0)
    x->bv.bv_self[index/CHAR_BIT] &= ~(0200>>index%CHAR_BIT);
  else
    x->bv.bv_self[index/CHAR_BIT] |= 0200>>index%CHAR_BIT;
  return value;
}

init_cmpaux()
{
	make_si_function("SPECIALP",siLspecialp);
}
