/*
    init.c  -- Lisp Initialization.
*/
/*
    Copyright (c) 1990, Giuseppe Attardi.

    ECoLisp is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    See file '../Copyright' for full details.
*/

#include "config.h"

/******************************* EXPORTS ******************************/

#ifndef RUNTIME
#define FUN_TABLE_SIZE 256
caddr_t	*function_entry_table;
int	function_entries_max = FUN_TABLE_SIZE;
int	function_entries = 0;
#endif

/******************************* ------- ******************************/

extern char system_directory[];

init_lisp()
{
	init_alloc();
#ifndef RUNTIME
	function_entry_table =
	  (caddr_t *)malloc(2 * function_entries_max * sizeof(caddr_t));
#endif
	init_symbol();
	init_package();
	init_typespec();
	init_number();
	init_character();
	init_file();
	init_read();
	init_print();
	init_bind();
	init_pathname();
	init_GC();
#ifdef unix
	init_load();
	init_unixsys();
#endif unix
	init_array();
	init_list();
	init_predicate();
	init_cfun();
	init_sequence();
	init_structure();
	init_string();
	init_alloc_function();
	init_character_function();
	init_file_function();
	init_package_function();
	init_pathname_function();
	init_print_function();
	init_read_function();
	init_symbol_function();
	init_typespec_function();
#ifdef TCP
	init_tcp();
#endif TCP
#ifdef THREADS
	init_lwp();
#endif THREADS
#ifdef CLOS
	init_instance();
	init_clos();
	init_gfun();
#endif CLOS
#ifdef TK
	init_tk();
#endif TK
#ifdef LOCATIVE
	init_unify();
#endif LOCATIVE
	init_hash();
#ifdef unix
	init_unixfsys();
	init_time();
#endif unix
	init_eval();
	init_lex();
	init_reference();
	init_assignment();
	init_error();
	init_toplevel();
	init_conditional();
	init_catch();
	init_macros();
	init_let();
	init_prog();
	init_block();
	init_multival();
	init_mapfun();
	init_iteration();
	init_cmpaux();
	init_main();
	init_format();
	init_interrupt();

	init_libs();
#ifdef COMPILER_IN_KERNEL
	init_compiler();
#endif COMPILER_IN_KERNEL
	SYM_VAL(Vpackage) = user_package;
#ifdef PROFILE
	init_profile();
#endif
}

static object
merge_system_directory(char *s)
{
	static char buff[MAXPATHLEN];
	extern char *strcat();

	strcpy(buff, system_directory);
	return(make_simple_string(strcat(buff, s)));
}

static object
open_input_stream(char *fn)
{
  return open_stream(make_simple_string(fn), smm_input, Cnil, Kerror);
}

init_libs()
{
	object stream;

	Lload(3, merge_system_directory("lsp/export.lsp"), Kverbose, Cnil);
	stream = open_input_stream("lsp/defmacro.data");
	init_defmacro(0, stream);
#ifndef RUNTIME
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/evalmacros.data");
	init_evalmacros(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/top.data");
	init_top(0, stream);
	close_stream(stream, TRUE);
#endif RUNTIME
	stream = open_input_stream("lsp/module.data");
	init_module(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/setf.data");
	init_setf(0, stream);
	close_stream(stream, TRUE);
	Lload(3, merge_system_directory("lsp/autoload.lsp"), Kverbose, Cnil);


	stream = open_input_stream("lsp/predlib.data");
	init_predlib(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/arraylib.data");
	init_arraylib(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/assert.data");
	init_assert(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/defstruct.data");
	init_defstruct(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/describe.data");
	init_describe(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/iolib.data");
	init_iolib(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/listlib.data");
	init_listlib(0, stream);
	close_stream(stream, TRUE);
#ifdef ANSI
	stream = open_input_stream("lsp/ansi.data");
	init_ansi(0, stream);
	close_stream(stream, TRUE);
/*
	stream = open_input_stream("lsp/loop.data");
	init_loop(0, stream);
	close_stream(stream, TRUE);
*/
#endif ANSI
	stream = open_input_stream("lsp/mislib.data");
	init_mislib(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/numlib.data");
	init_numlib(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/packlib.data");
	init_packlib(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/seq.data");
	init_seq(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/seqlib.data");
	init_seqlib(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("lsp/trace.data");
	init_trace(0, stream);
	close_stream(stream, TRUE);
#ifdef THREADS
	stream = open_input_stream("lsp/thread.data");
	init_thread(0, stream);
	close_stream(stream, TRUE);
#endif THREADS
#ifdef TK
	Lload(3, merge_system_directory("tk/tk-init.lsp"), Kverbose, Cnil);
#endif TK
      }

#ifdef COMPILER_IN_KERNEL
init_compiler()
{
	object stream;

	stream = open_input_stream("cmp/cmpdefs.data");
	init_cmpdefs(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpinline.data");
	init_cmpinline(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmputil.data");
	init_cmputil(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmptype.data");
	init_cmptype(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpbind.data");
	init_cmpbind(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpblock.data");
	init_cmpblock(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpcall.data");
	init_cmpcall(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpcatch.data");
	init_cmpcatch(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpenv.data");
	init_cmpenv(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpeval.data");
	init_cmpeval(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpflet.data");
	init_cmpflet(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpfun.data");
	init_cmpfun(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpif.data");
	init_cmpif(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpexit.data");
	init_cmpexit(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmplam.data");
	init_cmplam(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmplet.data");
	init_cmplet(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmploc.data");
	init_cmploc(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpmap.data");
	init_cmpmap(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpmulti.data");
	init_cmpmulti(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpspecial.data");
	init_cmpspecial(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmptag.data");
	init_cmptag(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmptop.data");
	init_cmptop(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpvar.data");
	init_cmpvar(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpwt.data");
	init_cmpwt(0, stream);
	close_stream(stream, TRUE);
	stream = open_input_stream("cmp/cmpmain.data");
	init_cmpmain(0, stream);
	close_stream(stream, TRUE);
}
#endif COMPILER_IN_KERNEL
