/*
    lwp.d -- Light weight processes.
*/
/*
    Copyright (c) 1990, Giuseppe Attardi.

    ECoLisp is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    See file '../Copyright' for full details.
*/


#include "config.h"

/******************************* EXPORTS ******************************/

lpd main_lpd;
lpd *clwp = &main_lpd;
int critical_level = 0;
pd *running_head;		/* front of running pd's  */
pd *running_tail;		/* back of running pd's   */
pd main_pd;

/******************************* IMPORTS ******************************/

extern scheduler_interruption; /* in unixint.c */
extern int writec_PRINTstream();
extern object readc();
extern gc();
extern enum type garbage_parameter;

/******************************* ------- ******************************/

#define thread_switch()  { setTimer(0); enable_scheduler(); \
			     scheduler(0, 0, NULL); }

static bool timer_active = FALSE;
static bool scheduler_disabled = FALSE;
static int scheduler_level = 0;            /* tito */
static bool reset_timer = FALSE;
static int running_processes = 1;
static int absolute_time = 0;

object Srunning;
object Ssuspended;
object Swaiting;
object Sstopped;
object Sdead;
object siSthread_top_level;

static object main_thread;

static
setTimer(long time)
{
  struct itimerval oldtimer;
  struct itimerval itimer;
  itimer.it_value.tv_sec = 0;
  itimer.it_value.tv_usec = time;
  itimer.it_interval.tv_sec = 0;
  itimer.it_interval.tv_usec = 0;
  setitimer(ITIMER_REAL, &itimer, &oldtimer);
}

pd *
dequeue()
{
  pd *tmp;
  tmp = running_head;
  if  (running_head != NULL)
    running_head = running_head->pd_next;
  return tmp;
}

pd *
make_pd()
{
  pd *new_pd; lpd *npd;
  
  /* Allocate a new descriptor for the new lwp */
  new_pd = (pd *)malloc(sizeof(pd));
  
  /* create a new stack ... */
  new_pd->pd_base = (int *)malloc(STACK_SIZE * sizeof(int));
  new_pd->pd_status = SUSPENDED;

  /* allocate a lisp descriptor:
   * using the calloc here it's possible to avoid the
   * critical section in the various push operations
   */
  npd = new_pd->pd_lpd = (lpd *)calloc(sizeof(lpd), 1);

  /* initialize it */

  /* bind stack */
  npd->lwp_bds_top = npd->lwp_bind_stack - 1;
  npd->lwp_bds_limit = npd->lwp_bind_stack + BDSSIZE;

  /* C stack */
  /* cssize is different now for the main thread only, but you might
     want to create threads with different stack sizes             */

#ifdef DOWN_STACK
  npd->lwp_cs_org = new_pd->pd_base + STACK_SIZE - 1;
  npd->lwp_cs_limit = new_pd->pd_base - 1;
#else
  npd->lwp_cs_org = new_pd->pd_base;
  npd->lwp_cs_limit = npd->lwp_cs_org + STACK_SIZE;
#endif    
  /* invocation history stack */
  npd->lwp_ihs_top = npd->lwp_ihs_stack - 1;
  npd->lwp_ihs_limit = npd->lwp_ihs_stack + IHSSIZE;
  /* frame stack */
  npd->lwp_frs_top = npd->lwp_frame_stack - 1;
  npd->lwp_frs_limit = npd->lwp_frame_stack + FRSSIZE;
  
  npd->lwp_alloc_temporary = OBJNULL;
  npd->lwp_backq_level = 0;
  npd->lwp_eval1 = 0;
  /* for gc */
  npd->lwp_fmt_temporary_stream = OBJNULL;
  npd->lwp_fmt_temporary_string = OBJNULL;
  
  npd->lwp_PRINTstream = Cnil;
  npd->lwp_PRINTescape = TRUE;
  npd->lwp_PRINTpretty = FALSE;
  npd->lwp_PRINTcircle = FALSE;
  npd->lwp_PRINTbase = 10;
  npd->lwp_PRINTradix = FALSE;
  npd->lwp_PRINTcase = Kdowncase;
  npd->lwp_PRINTgensym = TRUE;
  npd->lwp_PRINTlevel = -1;
  npd->lwp_PRINTlength = -1;
  npd->lwp_PRINTarray = FALSE;
  npd->lwp_write_ch_fun = writec_PRINTstream;
  npd->lwp_output_ch_fun = writec_PRINTstream;
  npd->lwp_read_ch_fun = readc;
  
  npd->lwp_READtable  = symbol_value(Vreadtable);
  npd->lwp_READdefault_float_format = 'S';
  npd->lwp_READbase = 10;
  npd->lwp_READsuppress = FALSE;
  npd->lwp_delimiting_char = OBJNULL;
  npd->lwp_detect_eos_flag = FALSE;
  npd->lwp_in_list_flag = FALSE;
  npd->lwp_dot_flag = FALSE;
  npd->lwp_sharp_eq_context_max = 0;
  
  /* for gc */
  npd->lwp_string_register = OBJNULL;
  npd->lwp_gensym_prefix = OBJNULL;
  npd->lwp_gentemp_prefix = OBJNULL;
  npd->lwp_token = OBJNULL;
    
  /* lex_env copy */
  npd->lwp_lex[0] = lex_env[0];
  npd->lwp_lex[1] = lex_env[1];
  npd->lwp_lex[2] = lex_env[2];
  npd->lwp_lex_env = npd->lwp_lex;
  
  /* ihs_push(Cnil) */
  (++npd->lwp_ihs_top)->ihs_function = Cnil;
  npd->lwp_ihs_top->ihs_base = npd->lwp_lex_env;

  /* Now the allocation. If the gc is invoked we are able to mark
   * the objects already allocated
   */
  npd->lwp_fmt_temporary_stream = make_string_output_stream(64);
  npd->lwp_fmt_temporary_string =
    npd->lwp_fmt_temporary_stream->sm.sm_object0;
  
  npd->lwp_string_register = alloc_simple_string(0);
  npd->lwp_gensym_prefix = make_simple_string("G");
  npd->lwp_gentemp_prefix = make_simple_string("T");
  npd->lwp_token = alloc_simple_string(LISP_PAGESIZE);
  npd->lwp_token->st.st_self = alloc_contblock(LISP_PAGESIZE);
  npd->lwp_token->st.st_fillp = 0;
  npd->lwp_token->st.st_hasfillp = TRUE;
  npd->lwp_token->st.st_adjustable = TRUE;
  
  return new_pd;
}

update_queue()
{
  register pd *dead_pd;
  pd *last = running_tail;

  do
    switch (running_head->pd_status) {

    case DEAD:

      /* remove the dead process */
      dead_pd = dequeue();
      /* free the lisp descriptor */
      free(dead_pd->pd_lpd);
      /* free the memory allocated for the stack and the descriptor */
      free(dead_pd->pd_base);
      free(dead_pd);
      break;

/*    case SUSPENDED: */
    case DELAYED:

      if (running_head->pd_slice != 0) 
	if (absolute_time > running_head->pd_slice) {

	  /* the time slice has expired */
	  running_head->pd_slice = 0;

	  if ((running_head->pd_thread->th.th_cont) != OBJNULL) {
	    /* in this case a continuation was created before %delay */
	    running_head->pd_thread->th.th_cont->cn.cn_timed_out = TRUE;
	    running_head->pd_thread->th.th_cont = OBJNULL;
	  }
	  running_head->pd_status = RUNNING;
	  return;		/* now you are a running process */
	}
      ROTQUEUE();
      break;
	  
    case WAITING:		/* waiting processes need to be scheduled  */
    case RUNNING:
      return;			/* found schedulable process */
	
    default:			/* currently is only STOPPED */
      ROTQUEUE();
      break;
    }
  while (running_head != last);
}

activate_thread(object thread)
{
  pd *npd = thread->th.th_self;

  /* jump on the new C stack */
  if (sigsetjmp(npd->pd_env, 1) == 0) {
#define STACK_MARGIN 160	/* longjmp writes also under the sp level */
#ifdef DOWN_STACK
# ifdef __linux
    npd->pd_env[0].__jmpbuf[0].__sp =
      stack_align((int)(npd->pd_base) + sizeof(int)*STACK_SIZE - STACK_MARGIN);
    npd->pd_lpd->lwp_cssize =
      npd->pd_env[0].__jmpbuf[0].__sp - (int)npd->pd_base;
# else
    npd->pd_env[JB_SP] =
      stack_align((int)(npd->pd_base) + sizeof(int)*STACK_SIZE - STACK_MARGIN);
    npd->pd_lpd->lwp_cssize =
      npd->pd_env[JB_SP] - (int)npd->pd_base;
# endif
#else
    npd->pd_env[JB_SP] =
      stack_align((int)(npd->pd_base));
    npd->pd_lpd->lwp_cssize = sizeof(int) * STACK_SIZE - STACK_MARGIN;
#endif DOWN_STACK
#ifdef JB_FP
    npd->pd_env[JB_FP] = npd->pd_env[JB_SP];
#endif
    return;
  }

  /* Back here when thread is first resumed.
   *
   * WARNING: args and locals are no more accessible from here on,
   * since we reenter with longjmp onto a new stack.
   */
  
#ifndef sun4sol2
  /* on SunOS sigmask is 0x82001 here */
  sigsetmask(sigblock(0) & ~(sigmask(SIGALRM)));
#endif

  { int i;
    for (i = clwp->lwp_nValues; i > 0; i--)
      /* gcc 2.7.2 optimizer on linux messes this:
	 VALUES(i) = VALUES(--i);
	 */
      VALUES(i) = VALUES(i-1);
    VALUES(0) = clwp->lwp_thread->th.th_fun;
    apply(clwp->lwp_nValues+1, siSthread_top_level, &VALUES(0));
  }
  /* Termination */
  
  terpri(Cnil);
  running_head->pd_status = DEAD;
  running_head->pd_thread->th.th_self = NULL;
  running_processes--;

  update_queue();
  thread_next();		/* update_queue has freed our stack!!! */
}

/*
 * switch to the first thread on queue
 */
thread_next()
{
  /* unwind the bind stack */
  lwp_bds_unwind(clwp->lwp_bind_stack, clwp->lwp_bds_top);

  /* switch clwp */
  clwp = running_head->pd_lpd;

  /* restore Values pointer */
  Values = clwp->lwp_Values;

  /* wind the bind stack */
  lwp_bds_wind(clwp->lwp_bind_stack, clwp->lwp_bds_top);

  /* reset the timer */
  if (running_processes > 1) {
    timer_active = TRUE;
    setTimer(REALQUANTUM); 
  } else { 
    timer_active = FALSE;
    absolute_time = 0;
  }
  siglongjmp(running_head->pd_env, 1); 
}

/*
 * Called when time slice expires or explicitily to switch thread
 */
scheduler(int sig, int code, struct sigcontext *scp)
{
  int val;

#if defined(SYSV) || defined(__svr4__) || defined(__linux)
  signal(SIGALRM, scheduler);
#endif SYSV

  absolute_time++;
  if (critical_level > 0) {	/* within critical section */
    scheduler_interrupted = TRUE;
    scheduler_interruption = SCHEDULER_INT;
    return;
  }
  if (scheduler_level > 0) {	/* abilitation  check */
    scheduler_interrupted = TRUE;
    return;
  }

  val = sigsetjmp(running_head->pd_env, 1);

  if (val == 1)			/* resume interrupted thread execution */
    return;			/* coming back from longjmp in thread_next */

  if (val == 2)			/* coming back from longjmp in GC */
    gc(garbage_parameter);	/* GC will return to the previous thread */

  ROTQUEUE();
  thread_next();
}

/*
 * Handle signal received within critical section
 */
interruption_handler()
{
  scheduler_interrupted = FALSE;

  switch (scheduler_interruption) {

  case SCHEDULER_INT:
    thread_switch();
    break;

  case ERROR_INT:
    sigint();
    break;
  }
}

lwp_bds_wind(bds_ptr base, bds_ptr top)
{
  object temp;

  for (; top >= base; base++) {
    temp = SYM_VAL(base->bds_sym);
    SYM_VAL(base->bds_sym) = base->bds_val;
    base->bds_val = temp;
  }
}

lwp_bds_unwind(bds_ptr base, bds_ptr top)
{
  object temp;

  for (; top >= base; top--) {
    temp = SYM_VAL(top->bds_sym);
    SYM_VAL(top->bds_sym) = top->bds_val;
    top->bds_val = temp;
  }
}

resume(pd *rpd)
{
  register pd *p;

  start_critical_section();
  running_processes++;

  rpd->pd_status = RUNNING;
  for (p = running_head; (p != rpd) && (p != NULL); p = p->pd_next) ;
  if (p == NULL)
    ENQUEUE(rpd);
  end_critical_section();

  if (!timer_active)  {
    timer_active = TRUE;
    setTimer(REALQUANTUM);
  }
}

/***********
 *
 * THREADS
 *
 ***********/


siLthread_break_in(int narg)
{
  alarm(0);
  VALUES(0) = Cnil;
  RETURN(1);
}

siLthread_break_quit(int narg)
{
  /* reset everything in MT */
  pd *p;

  /* this is done in any case to remedy the problem with C-c handling */
  signal(SIGALRM, scheduler);

  if (timer_active) {
     /* reset the critical and disable-scheduler environment    */
      scheduler_disabled = FALSE;
      scheduler_level = 0;
      critical_level = 0;
      scheduler_interrupted = 0;

      for (p = running_head; (p != NULL); p = p->pd_next)
	if (p != &main_pd)
	  p->pd_status = DEAD;
	else {
	  p->pd_status = RUNNING;
	  p->pd_thread->th.th_cont = OBJNULL;
	}

      if (running_head != &main_pd) {
	update_queue();
	thread_next(); 
	/*  here one should deallocate the main-thread function     */
      }
      else
	thread_switch();
    }
  VALUES(0) = Cnil;
  RETURN(1);
}

siLthread_break_resume(int narg)
{
  /* Restart the timer that might have been
   * changed by the interrupt handling
   */
  signal(SIGALRM, scheduler);  
  if (timer_active)
    thread_switch();
  VALUES(0) = Cnil;
  RETURN(1);
}

Lthread_list(int narg)
{
  pd *p;
  object tmp, x = CONS(running_head->pd_thread, Cnil);

  tmp = x;

  start_critical_section();

  for (p = running_head->pd_next; (p != NULL); p = p->pd_next) {
      CDR(tmp) = CONS(p->pd_thread, Cnil);
      tmp = CDR(tmp);
    }

  end_critical_section();

  VALUES(0) = x;
  RETURN(1);
}

Lmake_thread(int narg, object fun)
{
  object x;
  pd *npd;
  /* Just one argument for the time being */
  check_arg(1);

  if (type_of(fun) == t_symbol) {
    if (fun->s.s_sfdef != NOT_SPECIAL ||
	fun->s.s_mflag)
      FEinvalid_function(fun);
    if (SYM_FUN(fun) == OBJNULL)
      FEundefined_function(fun);
    /*  fun = SYM_FUN(fun); confusing */
  }

  x = alloc_object(t_thread);
  x->th.th_fun = fun;
  x->th.th_size = sizeof(pd);
  x->th.th_self = npd = make_pd();
  x->th.th_cont = OBJNULL;

  npd->pd_thread = x;
  npd->pd_slice = 0;

  /* Backpointer to thread */
  npd->pd_lpd->lwp_thread = x;

  activate_thread(x);

  VALUES(0) = x;
  RETURN(1);
}

Ldeactivate(int narg, object thread)
{
  check_arg(1);
  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_self == NULL ||
      thread->th.th_self->pd_status != RUNNING)
    FEerror("Cannot deactivate a thread not running", 0);

  start_critical_section();	/* tito */
  thread->th.th_self->pd_status = STOPPED;
  running_processes--;
  if (thread->th.th_self == running_head) {
    critical_level--; /* end_critical_section() */
    update_queue();
    thread_next();
  } else
    end_critical_section();
  VALUES(0) = Cnil;
  RETURN(1);
}

Lreactivate(int narg, object thread)
{
  check_arg(1);
  start_critical_section();

  if (type_of(thread) != t_thread) {
    FEwrong_type_argument(Sthread, thread);
  }

  if (thread->th.th_self == NULL ||
      thread->th.th_self->pd_status != STOPPED)
    FEerror("Cannot reactivate a thread not stopped", 0);

  start_critical_section();	/* tito */
  thread->th.th_self->pd_status = RUNNING;
  running_processes++;

  if (!timer_active) {
     timer_active = TRUE;
     setTimer(REALQUANTUM);
   }

  end_critical_section();
  VALUES(0) = Cnil;
  RETURN(1);

}

Lkill_thread(int narg, object thread)
{

  /* The following code is not enough.
     Consider: The scheduler can be disabled
     What about killing the current thread?
     */
  check_arg(1);

  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_self != NULL) {
    start_critical_section();
    thread->th.th_self->pd_status = DEAD;
    if (thread->th.th_self->pd_lpd == clwp) {
       /* if a thread kills itself the scheduler is to be called */
       thread->th.th_self = NULL;
       critical_level--; /*  end_critical_section() */
       update_queue();
       thread_next();
     }
     else {
       thread->th.th_self = NULL;
       end_critical_section();
     }
  }
  VALUES(0) = Cnil;
  RETURN(1);
}

Lcurrent_thread(int narg)
{
  check_arg(0);
  VALUES(0) = clwp->lwp_thread;
  RETURN(1);
}

Lthread_status(int narg, object thread)
{
  check_arg(1);

  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_self != NULL)
    switch (thread->th.th_self->pd_status) {
    case RUNNING:
      VALUES(0) = Srunning;
      break;
    case SUSPENDED:
      VALUES(0) = Ssuspended;
      break;
    case WAITING:
      VALUES(0) = Swaiting;
      break;
    case STOPPED:
      VALUES(0) = Sstopped;
      break;
    case DEAD:
      VALUES(0) = Sdead;
      break;
    default:
      FEerror("Unexpected type for thread ~A", 1, thread);
    }
  else
    VALUES(0) = Sdead;

  RETURN(1);
}


/***************
 *
 * CONTINUATIONS
 *
 ***************/

Lmake_continuation(int narg, object thread)
{
  object x;
  check_arg(1);

  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);

  if (thread->th.th_cont)
    FEerror("A continuation for thread ~A already exists.", 1, thread);

  if (thread->th.th_self == NULL ||
      thread->th.th_self->pd_status == DEAD) {
    FEerror("Thread ~A is DEAD.", 1, thread);
  }

  x = alloc_object(t_cont);

  x->cn.cn_thread = thread;
  x->cn.cn_resumed = FALSE;
  x->cn.cn_timed_out = FALSE;

  VALUES(0) = thread->th.th_cont = x;
  RETURN(1);
}

/* Returns the thread associated to a continuation */
Lthread_of(int narg, object cont)
{
  check_arg(1);
  if (type_of(cont) != t_cont)
    FEwrong_type_argument(Scont, cont);
  VALUES(0) = cont->cn.cn_thread;
  RETURN(1);
}

/* Returns the continuation associated to a thread, if it exists */
Lcontinuation_of(int narg, object thread)
{
  check_arg(1);
  if (type_of(thread) != t_thread)
    FEwrong_type_argument(Sthread, thread);
  if (thread->th.th_cont)
    VALUES(0) = thread->th.th_cont;
  else 
    VALUES(0) = Cnil;
  RETURN(1);
}

Lresume(int narg, object cont, ...)
{
  int i;
  va_list args;
  object *thread_Values;

  if (narg < 1) FEtoo_few_arguments(&narg);

  if (Null(cont)) {
    VALUES(0) = Cnil;
    RETURN(1);
  }

  if (type_of(cont) != t_cont)
    FEwrong_type_argument(Scont, cont);

  if (cont->cn.cn_resumed)
    FEerror("The continuation has already been resumed.", 0);
  
  if (cont->cn.cn_timed_out)
    FEerror("The continuation has been timed out.", 0);

  if (cont->cn.cn_thread->th.th_self == NULL)
    FEerror("The continuation cannot be resumed. Its thread is DEAD.", 0);

  if (cont->cn.cn_thread->th.th_self->pd_status != SUSPENDED &&
      cont->cn.cn_thread->th.th_self->pd_status != DELAYED)
    FEerror("The continuation cannot be resumed. Its thread isn't suspended", 0);

  /* Push the arguments on the value stack of thread */

  thread_Values = cont->cn.cn_thread->th.th_self->pd_lpd->lwp_Values;

  va_start(args, cont);
  for (i = 1; i < narg; i++)
    *(thread_Values++) = va_arg(args, object);
  cont->cn.cn_thread->th.th_self->pd_lpd->lwp_nValues = narg-1;

  cont->cn.cn_resumed = TRUE;
  cont->cn.cn_thread->th.th_cont = OBJNULL;

  /* If you are waiting on a slice expiring I reset your slice    */
  cont->cn.cn_thread->th.th_self->pd_slice = 0;

  resume(cont->cn.cn_thread->th.th_self);

  VALUES(0) = cont->cn.cn_thread;
  RETURN(1);
}


/***************
 *
 * SCHEDULING
 *
 ***************/


Ldisable_scheduler(int narg)
{
  check_arg(0);

  scheduler_level++;
  VALUES(0) = Cnil;
  RETURN(1);
}

Lenable_scheduler(int narg)
{
  check_arg(0);

  enable_scheduler();
  VALUES(0) = Cnil;
  RETURN(1);
}

enable_scheduler()
{
   scheduler_level = 0;

   if (scheduler_interrupted) {
     scheduler_interrupted = FALSE;
     thread_switch();
   }
}

Lsuspend(int narg)
{
  check_arg(0);

  if (timer_active) {
       running_head->pd_status = SUSPENDED;
       running_processes--;
       thread_switch();
       /* When resumed it will be provided with the Values to return */
       RETURN(running_head->pd_lpd->lwp_nValues);
     }
  else 
    FEerror("No other active thread.", 0);
}

Ldelay(int narg, object interval)
{ int z;
  
  check_arg(1);
  check_type_non_negative_integer(&interval);
  z = fix(interval);
  
  if (timer_active) {
    running_head->pd_status = DELAYED;
    running_processes--;
       
    /* Translate seconds in number of scheduler slices    */
    running_head->pd_slice = z * 10 + absolute_time;

    thread_switch();
       
    /* When resumed it will be provided with the Values to return */
    RETURN(running_head->pd_lpd->lwp_nValues);
  }
  else
    sleep(z);
}

Lthread_wait(int narg, object fun, ...)
{ int nr;
  va_list args;
  va_start(args, fun);

  if (narg < 1) FEtoo_few_arguments(&narg);

  start_critical_section();
  running_head->pd_status = WAITING;
  running_processes--;
  end_critical_section();

  for (;;) {
    
    nr = apply(narg-1, fun, args);
    
    if (VALUES(0) != Cnil)
      break;
    else if (timer_active) {
      /* the time slice has not been used */
      absolute_time--;
      thread_switch();
    } else
      FEerror("The condition will never be satisfied for lack of active processes", 0);
  }
  running_head->pd_status = RUNNING;
  running_processes++;
  end_critical_section();
  RETURN(nr);
}
  
      
Lthread_wait_with_timeout(int narg, object timeout, object fun, ...)
{
  int nr;
  va_list args;
  va_start(args, fun);

  if (narg < 2) FEtoo_few_arguments(&narg);
  check_type_non_negative_integer(&timeout);

  /* We have to translate seconds in scheduler call number */
  start_critical_section();
  running_head->pd_slice = fix(timeout) * 10 + absolute_time;
  
  running_head->pd_status = WAITING;
  running_processes--;
  end_critical_section();
  
  for (;;) {
    
    if (absolute_time > running_head->pd_slice) {
      /* the time slice has expired */
      VALUES(0) = Cnil;
      nr = 1;
      break;
    }

    nr = apply(narg-1, fun, args);

    if (VALUES(0) != Cnil)
      break;
    else {
      /* the time slice has not been used */
      absolute_time--;
      thread_switch();
    }
  }
  
  start_critical_section();
  running_head->pd_slice = 0;
  running_head->pd_status = RUNNING;
  running_processes++;
  end_critical_section();
  RETURN(nr);
}

enable_lwp()
{
  signal(SIGALRM, scheduler);
}

init_lwp()
{ pd *temp_pd;

  temp_pd = &main_pd;
  PUSH(temp_pd);

  main_thread = alloc_object(t_thread);
  main_pd.pd_thread = main_thread;
  main_thread->th.th_fun = Cnil;
  main_thread->th.th_size = sizeof (pd);
  main_thread->th.th_self = &main_pd;
  main_thread->th.th_cont = OBJNULL;
  /* Backpointer to thread */
  main_pd.pd_status = RUNNING;
  main_pd.pd_lpd = &main_lpd;
  main_lpd.lwp_thread = main_thread;
  register_root(&main_thread);

  Srunning = make_ordinary("RUNNING");
  Ssuspended = make_ordinary("SUSPENDED");
  Swaiting = make_ordinary("WAITING");
  Sstopped = make_ordinary("STOPPED");
  Sdead = make_ordinary("DEAD");
  siSthread_top_level = make_si_ordinary("THREAD-TOP-LEVEL");

  make_si_function("THREAD-BREAK-IN", siLthread_break_in);
  make_si_function("THREAD-BREAK-QUIT", siLthread_break_quit);
  make_si_function("THREAD-BREAK-RESUME", siLthread_break_resume);

  make_function("MAKE-THREAD", Lmake_thread);
  make_function("DEACTIVATE", Ldeactivate);
  make_function("REACTIVATE", Lreactivate);
  make_function("KILL-THREAD", Lkill_thread);
  
  make_function("CURRENT-THREAD", Lcurrent_thread);
  make_function("THREAD-STATUS", Lthread_status);
  make_function("THREAD-LIST", Lthread_list);
  
  make_function("MAKE-CONTINUATION", Lmake_continuation);
  make_function("THREAD-OF", Lthread_of);
  make_function("CONTINUATION-OF", Lcontinuation_of);
  make_function("RESUME", Lresume);

  make_function("%DISABLE-SCHEDULER", Ldisable_scheduler);
  make_function("%ENABLE-SCHEDULER", Lenable_scheduler);

  make_function("%SUSPEND", Lsuspend);
  make_function("%DELAY", Ldelay);
  make_function("%THREAD-WAIT", Lthread_wait);
  make_function("%THREAD-WAIT-WITH-TIMEOUT", Lthread_wait_with_timeout);
}
