/* tcp.c  -- stream interface to TCP					*/

/*
    Copyright (c) 1990, Giuseppe Attardi.

    ECoLisp is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Library Public License as published
    by the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    See file '../Copyright' for full details.
*/

#include "config.h"

object
make_stream(object host, int fd, enum smmode smm)
{
   object stream;
   char *mode;			/* file open mode */
   FILE *fp;			/* file pointer */

   switch(smm) {
    case smm_io:
      mode = "r+";
      break;
    case smm_input:
      mode = "r";
      break;
    case smm_output:
      mode = "w";
      break;
    default:
      FEerror("make_stream: wrong mode", 0);
   }
   fp = fdopen(fd, mode);

   stream = alloc_object(t_stream);
   stream->sm.sm_mode = (short)smm;
   stream->sm.sm_fp = fp;
   fp->_IO_buf_base = NULL; /* BASEFF */; 
   stream->sm.sm_object0 = Sstring_char;
   stream->sm.sm_object1 = host; /* not really used */
   stream->sm.sm_int0 = stream->sm.sm_int1 = 0;
   setbuf(fp, alloc_contblock(BUFSIZ)); 
   return(stream);
}

/* 
   Lopen_client_stream --

   To test this function, try:
   (setq s (si:open-client-stream "host" 13))
   (read-line s)
   "Wed Jun 22 19:44:36 METDST 1994"
*/
Lopen_client_stream(int narg, object host, object port)
{
   int fd;			/* file descriptor */
   object stream;

   check_arg(2);
   if (type_of(host) != t_string)
     FEwrong_type_argument(Sstring, host);

   if (!FIXNUMP(port))
     FEwrong_type_argument(TSpositive_number, port);

   if (host->st.st_fillp > BUFSIZ - 1)
     FEerror("~S is a too long file name.", 1, host);

#ifdef THREADS
   start_critical_section();
#endif THREADS
   fd = connect_to_server(host->st.st_self, fix(port)); 
#ifdef THREADS
   end_critical_section();
#endif THREADS

   if (fd == 0) {
     VALUES(0) = Cnil;
     RETURN(1);}

   stream = make_stream(host, fd, smm_io);

   VALUES(0) = make_two_way_stream(stream, stream);
   RETURN(1);
}

object
Lopen_server_stream(int narg, object port)
{  int fd;			/* file descriptor */
   object stream;

   check_arg(1);
   if (!FIXNUMP(port))
     FEwrong_type_argument(TSpositive_number, port);

#ifdef THREADS
   start_critical_section();
#endif THREADS
   fd = create_server_port(fix(port));
#ifdef THREADS
   end_critical_section();
#endif THREADS

   if (fd == 0)
     VALUES(0) = Cnil;
   else {
     stream = make_stream(Cnil, fd, smm_io);
     VALUES(0) = make_two_way_stream(stream, stream);
   }
   RETURN(1);
}

init_tcp()
{
  make_si_function("OPEN-CLIENT-STREAM", Lopen_client_stream);
  make_si_function("OPEN-SERVER-STREAM", Lopen_server_stream);
}
