/*
    unixsys.s  -- Unix shell interface.
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.

    ECoLisp is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    See file '../Copyright' for full details.
*/

#include "config.h"

#if !defined(__stdlib_h) && !defined(_STDLIB_H_) && !defined(__STDLIB_H__) &&  !defined(_STDLIB_H)
#include <signal.h>
int
system(const char *command)
{
	char buf[4];
	extern sigint();

	signal(SIGINT, SIG_IGN);
	write(4, command, strlen(command)+1);
	read(5, buf, 1);
	signal(SIGINT, sigint);
	return(buf[0]<<8);
}
#endif __STDLIB_H__

#ifdef __FreeBSD__

/* due to the calls to realloc in system.c/exec.c (memory which hasn't been
malloc'ed can't be realloced in ecl) we have to patch this a bit.
We use execv and supply the arg list, so execl doesn't have to realloc. CvdL */

#include <sys/types.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <paths.h>

system(command)
	const char *command;
{
	union wait pstat;
	pid_t pid;
	int omask;
	sig_t intsave, quitsave;

	if (!command)		/* just checking... */
		return(1);

	omask = sigblock(sigmask(SIGCHLD));
	switch(pid = vfork()) {
	case -1:			/* error */
		(void)sigsetmask(omask);
		pstat.w_status = 0;
		pstat.w_retcode = 127;
		return(pstat.w_status);
	case 0:				/* child */
	  { char *args[] = { {"sh"}, {"-c"}, {command}, {(char *)NULL} };
		(void)sigsetmask(omask);
	        execv(_PATH_BSHELL, args);
		_exit(127);
	  }
	}
	intsave = signal(SIGINT, SIG_IGN);
	quitsave = signal(SIGQUIT, SIG_IGN);
	pid = waitpid(pid, (int *)&pstat, 0);
	(void)sigsetmask(omask);
	(void)signal(SIGINT, intsave);
	(void)signal(SIGQUIT, quitsave);
	return(pid == -1 ? -1 : pstat.w_status);
}
#endif

Lsystem(int narg, object cmd)
{
	char command[1024];
	int i;

	check_arg(1);
	check_type_string(&cmd);
	if (cmd->st.st_fillp >= 1024)
		FEerror("Too long command line: ~S.", 1, cmd);
	VALUES(0) = MAKE_FIXNUM(system(cmd->st.st_self) >> 8);
	RETURN(1);
}

siLopen_pipe(int narg, object cmd)
{
  FILE *ptr;
  object stream;

  check_arg(1);
  check_type_string(&cmd);
 
  if ((ptr = popen(cmd->st.st_self, OPEN_R)) == NULL) {
    VALUES(0) = Cnil;
    RETURN(1);
  }
  stream = alloc_object(t_stream);
  stream->sm.sm_mode = smm_input;
  stream->sm.sm_fp = ptr;
  ptr->_IO_buf_base = NULL;
  stream->sm.sm_object0 = Sstring_char;
  stream->sm.sm_object1 = cmd;
  stream->sm.sm_int0 = stream->sm.sm_int1 = 0;
  setbuf(ptr, alloc_contblock(BUFSIZ)); 
  VALUES(0) = stream;
  return(1);
}
 
init_unixsys()
{
	make_function("SYSTEM", Lsystem);
	make_si_function("OPEN-PIPE", siLopen_pipe);
}
